import sys
import unittest

from pathlib import Path
import tempfile
import networkx as nx
from functools import partial
from collections import Counter

from alpaca import ProvenanceGraph, alpaca_setting


class ProvenanceGraphTestCase(unittest.TestCase):

    @staticmethod
    def _attr_comparison(attr_G1, attr_G2):
        return attr_G1 == attr_G2

    @staticmethod
    def _gexf_edge_comparison(attr_G1, attr_G2):
        # GEXF files add an `id` attribute to the edge, ignore it
        if 'id' in attr_G1:
            attr_G1.pop('id')
        if 'id' in attr_G2:
            attr_G2.pop('id')
        return attr_G1 == attr_G2

    @classmethod
    def setUpClass(cls):
        cls.ttl_path = Path(__file__).parent / "res"
        cls.temp_dir = tempfile.TemporaryDirectory(dir=cls.ttl_path,
                                                   suffix="tmp")
        cls.graph_comparison = partial(nx.is_isomorphic,
                                       node_match=cls._attr_comparison)
        alpaca_setting('authority', "my-authority")

    def test_graph_behavior_and_serialization(self):
        input_file = self.ttl_path / "input_output.ttl"
        graph = ProvenanceGraph(input_file)
        self.assertIsInstance(graph.graph, nx.DiGraph)
        self.assertEqual(len(graph.graph.nodes), 3)
        self.assertEqual(len(graph.graph.edges), 2)

        output_path = Path(self.temp_dir.name)

        # Test GEXF serialization
        gexf_file = output_path / "test.gexf"
        graph.save_gexf(gexf_file)
        self.assertTrue(gexf_file.exists())
        read_gexf = nx.read_gexf(gexf_file)
        self.assertTrue(self.graph_comparison(
            read_gexf, graph.graph, edge_match=self._gexf_edge_comparison))

        # Test GraphML serialization
        graphml_file = output_path / "test.graphml"
        graph.save_graphml(graphml_file)
        self.assertTrue(graphml_file.exists())
        read_graphml = nx.read_graphml(graphml_file)
        self.assertTrue(self.graph_comparison(read_graphml, graph.graph,
                                              edge_match=self._attr_comparison))

    def test_use_name_in_parameter(self):
        node = "urn:fz-juelich.de:alpaca:function_execution:Python:111111:999999:test.test_function#12345"
        input_file = self.ttl_path / "input_output.ttl"
        graph_use_name = ProvenanceGraph(input_file)
        graph_dont_use_name = ProvenanceGraph(input_file,
                                              use_name_in_parameter=False)
        use_name_attrs = graph_use_name.graph.nodes[node]
        dont_use_name_attrs = graph_dont_use_name.graph.nodes[node]

        self.assertTrue('test_function:param_1' in use_name_attrs)
        self.assertTrue('parameter:param_1' in dont_use_name_attrs)
        self.assertFalse('parameter:param_1' in use_name_attrs)
        self.assertFalse('test_function:param_1' in dont_use_name_attrs)
        self.assertEqual(use_name_attrs['test_function:param_1'], '5')
        self.assertEqual(dont_use_name_attrs['parameter:param_1'], '5')

    def test_remove_none(self):
        node = "urn:fz-juelich.de:alpaca:object:Python:builtins.NoneType:777777"
        input_file = self.ttl_path / "file_output.ttl"
        graph_with_none = ProvenanceGraph(input_file, remove_none=False)
        graph_without_none = ProvenanceGraph(input_file, remove_none=True)

        self.assertTrue(node in graph_with_none.graph.nodes)
        self.assertFalse(node in graph_without_none.graph.nodes)
        self.assertEqual(len(graph_with_none.graph.nodes), 4)
        self.assertEqual(len(graph_without_none.graph.nodes), 3)

    def test_remove_none_no_output_function(self):
        node = "urn:fz-juelich.de:alpaca:object:Python:builtins.NoneType:777777"
        input_file = self.ttl_path / "none_output.ttl"
        graph_with_none = ProvenanceGraph(input_file, remove_none=False)
        graph_without_none = ProvenanceGraph(input_file, remove_none=True)

        self.assertTrue(node in graph_with_none.graph.nodes)
        self.assertFalse(node in graph_without_none.graph.nodes)
        self.assertEqual(len(graph_with_none.graph.nodes), 3)
        self.assertEqual(len(graph_without_none.graph.nodes), 2)

    def test_memberships(self):
        input_file = self.ttl_path / "multiple_memberships.ttl"
        graph = ProvenanceGraph(input_file)

        self.assertEqual(len(graph.graph.nodes), 7)
        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
        "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
        "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345",
        "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333"):
            self.assertTrue(node in graph.graph.nodes)

        for expected_edge, expected_label in zip(
                ((
                 "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
                 "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332"),
                 (
                 "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
                 "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"),
                 (
                 "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
                 "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332"),
                 (
                 "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
                 "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")),
                (".containers", "[0]", ".inputs", "[1]")):
            self.assertTrue(expected_edge in graph.graph.edges)
            self.assertTrue(graph.graph.edges[expected_edge]['membership'])
            self.assertEqual(graph.graph.edges[expected_edge]['label'],
                             expected_label)

    def test_membership_condensing(self):
        input_file = self.ttl_path / "multiple_memberships.ttl"
        graph = ProvenanceGraph(input_file)
        graph.condense_memberships()

        self.assertEqual(len(graph.graph.nodes), 4)

        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345",
        "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333"):
            self.assertTrue(node in graph.graph.nodes)

        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
        "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"):
            self.assertFalse(node in graph.graph.nodes)

        for edge in ((
                     "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
                     "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
                     "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")):
            self.assertFalse(edge in graph.graph.edges)

        expected_edge = (
        "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
        "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")

        expected_label = ".containers[0].inputs[1]"

        self.assertTrue(expected_edge in graph.graph.edges)
        self.assertTrue(graph.graph.edges[expected_edge]['membership'])
        self.assertEqual(graph.graph.edges[expected_edge]['label'],
                         expected_label)

    def test_membership_condensing_with_preservation(self):
        input_file = self.ttl_path / "multiple_memberships.ttl"
        graph = ProvenanceGraph(input_file)
        graph.condense_memberships(preserve=['Container'])

        self.assertEqual(len(graph.graph.nodes), 5)

        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345",
        "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
        "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"):
            self.assertTrue(node in graph.graph.nodes)

        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332"):
            self.assertFalse(node in graph.graph.nodes)

        for edge in ((
                     "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
                     "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
                     "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")):
            self.assertFalse(edge in graph.graph.edges)

        for expected_edge, expected_label in zip(
                ((
                 "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
                 "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"),
                 (
                 "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
                 "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")),
                (".containers[0]", ".inputs[1]")):
            self.assertTrue(expected_edge in graph.graph.edges)
            self.assertTrue(graph.graph.edges[expected_edge]['membership'])
            self.assertEqual(graph.graph.edges[expected_edge]['label'],
                             expected_label)

    def test_strip_namespace(self):
        input_file = self.ttl_path / "metadata.ttl"

        graph = ProvenanceGraph(input_file, attributes=['name'],
                                annotations=['sua'], strip_namespace=True)
        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"]

        self.assertEqual(node_attrs["name"], "Spiketrain#1")
        self.assertEqual(node_attrs["sua"], "false")

    def test_use_class_in_method_name(self):
        input_file = self.ttl_path / "class_method.ttl"

        graph = ProvenanceGraph(input_file, attributes=None,
                                annotations=None,
                                use_class_in_method_name=True)
        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:function_execution:Python:111111:999999:test.ObjectWithMethod.process#12345"]

        self.assertEqual(node_attrs["label"], "ObjectWithMethod.process")

    def test_no_use_class_in_method_name(self):
        input_file = self.ttl_path / "class_method.ttl"

        graph = ProvenanceGraph(input_file, attributes=None,
                                annotations=None,
                                use_class_in_method_name=False)
        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:function_execution:Python:111111:999999:test.ObjectWithMethod.process#12345"]

        self.assertEqual(node_attrs["label"], "process")

    def test_no_strip_namespace(self):
        input_file = self.ttl_path / "metadata.ttl"

        graph = ProvenanceGraph(input_file, attributes=['name'],
                                annotations=['sua'], strip_namespace=False)
        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"]

        self.assertEqual(node_attrs["attribute:name"], "Spiketrain#1")
        self.assertEqual(node_attrs["annotation:sua"], "false")

    def test_attributes(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, attributes=['metadata_2'])

        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345"]

        self.assertEqual(node_attrs["metadata_2"], "5")

        for annotation in ("metadata_1", "metadata_3", "metadata_4"):
            self.assertTrue(annotation not in node_attrs)

        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"]

        self.assertTrue("name" not in node_attrs)

    def test_annotations(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, annotations=['sua'])

        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"]

        self.assertEqual(node_attrs['sua'], "false")

        for annotation in ("channel", "complexity", "event"):
            self.assertTrue(annotation not in node_attrs)

    def test_all_annotations(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, annotations='all')

        annotations_node = "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"

        expected_annotations = {"sua": "false",
                                "channel": "56",
                                "complexity": "[0 1 2 3]",
                                "event": "[ True False False]"}

        attributes_node = "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345"

        expected_attributes = {"metadata_1": "value1",
                               "metadata_2": "5",
                               "metadata_3": "5.0",
                               "metadata_4": "true"}

        node_attrs = graph.graph.nodes[annotations_node]
        for key, value in expected_annotations.items():
            self.assertTrue(key in node_attrs)
            self.assertEqual(node_attrs[key], value)

        node_attrs = graph.graph.nodes[attributes_node]
        for key, value in expected_attributes.items():
            self.assertTrue(key not in node_attrs)

    def test_all_attributes(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, attributes='all')

        annotations_node = "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"

        expected_annotations = {"sua": "false",
                                "channel": "56",
                                "complexity": "[0 1 2 3]",
                                "event": "[ True False False]"}

        attributes_node = "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345"

        expected_attributes = {"metadata_1": "value1",
                               "metadata_2": "5",
                               "metadata_3": "5.0",
                               "metadata_4": "true"}

        node_attrs = graph.graph.nodes[annotations_node]
        for key, value in expected_annotations.items():
            self.assertTrue(key not in node_attrs)

        self.assertEqual(node_attrs['name'], "Spiketrain#1")

        node_attrs = graph.graph.nodes[attributes_node]
        for key, value in expected_attributes.items():
            self.assertTrue(key in node_attrs)
            self.assertEqual(node_attrs[key], value)

    def test_remove_multiple_attributes(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, attributes='all',
                                annotations='all')

        graph.remove_attributes('metadata_4', 'sua', 'Time Interval')

        annotations_node = "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"

        expected_annotations = {"channel": "56",
                                "complexity": "[0 1 2 3]",
                                "event": "[ True False False]"}

        attributes_node = "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345"

        expected_attributes = {"metadata_1": "value1",
                               "metadata_2": "5",
                               "metadata_3": "5.0"}

        node_attrs = graph.graph.nodes[annotations_node]
        for key, value in expected_annotations.items():
            self.assertTrue(key in node_attrs)
            self.assertEqual(node_attrs[key], value)

        self.assertEqual(node_attrs['name'], "Spiketrain#1")

        node_attrs = graph.graph.nodes[attributes_node]
        for key, value in expected_attributes.items():
            self.assertTrue(key in node_attrs)
            self.assertEqual(node_attrs[key], value)

        for _, node_attrs in graph.graph.nodes(data=True):
            self.assertTrue("Time Interval" not in node_attrs)
            self.assertTrue("sua" not in node_attrs)
            self.assertTrue("metadata_4" not in node_attrs)

    def test_remove_attributes(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, attributes='all',
                                annotations='all')

        graph.remove_attributes('metadata_4')

        annotations_node = "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"

        expected_annotations = {"sua": "false",
                                "channel": "56",
                                "complexity": "[0 1 2 3]",
                                "event": "[ True False False]"}

        attributes_node = "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345"

        expected_attributes = {"metadata_1": "value1",
                               "metadata_2": "5",
                               "metadata_3": "5.0"}

        node_attrs = graph.graph.nodes[annotations_node]
        for key, value in expected_annotations.items():
            self.assertTrue(key in node_attrs)
            self.assertEqual(node_attrs[key], value)

        self.assertEqual(node_attrs['name'], "Spiketrain#1")

        node_attrs = graph.graph.nodes[attributes_node]
        for key, value in expected_attributes.items():
            self.assertTrue(key in node_attrs)
            self.assertEqual(node_attrs[key], value)

        for _, node_attrs in graph.graph.nodes(data=True):
            self.assertTrue("Time Interval" in node_attrs)
            self.assertTrue("metadata_4" not in node_attrs)

    def test_remove_attributes_aggregation(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, attributes='all',
                                annotations='all')
        aggregated = graph.aggregate({},
                                     remove_attributes="Time Interval")

        for _, node_attrs in aggregated.nodes(data=True):
            self.assertTrue("Time Interval" not in node_attrs)

    def test_remove_multiple_attributes_aggregation(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, attributes='all',
                                annotations='all')
        aggregated = graph.aggregate({},
                                     remove_attributes=("Time Interval",
                                                        "sua"))

        for _, node_attrs in aggregated.nodes(data=True):
            self.assertTrue("Time Interval" not in node_attrs)
            self.assertTrue("sua" not in node_attrs)

    def test_value_attribute(self):
        input_file = self.ttl_path / "values.ttl"
        graph = ProvenanceGraph(input_file, attributes='all',
                                annotations='all', value_attribute='value')

        node_values_by_id = {
            "urn:fz-juelich.de:alpaca:object:Python:builtins.int:543211": 1,
            "urn:fz-juelich.de:alpaca:object:Python:builtins.float:543212": 1.1,
            "urn:fz-juelich.de:alpaca:object:Python:builtins.str:543213": "test",
            "urn:fz-juelich.de:alpaca:object:Python:builtins.complex:543214": "(3+5j)",
            "urn:fz-juelich.de:alpaca:object:Python:builtins.bool:543215": True,
            "urn:fz-juelich.de:alpaca:object:Python:numpy.float32:543216": 1.2,
            "urn:fz-juelich.de:alpaca:object:Python:numpy.float64:543217": 1.3,
            "urn:fz-juelich.de:alpaca:object:Python:numpy.int64:543218": 2,
            "urn:fz-juelich.de:alpaca:object:Python:numpy.int32:543219": 3,
            "urn:fz-juelich.de:alpaca:object:Python:numpy.int16:5432110": -4,
            "urn:fz-juelich.de:alpaca:object:Python:builtins.dict:5432111": "{'id': [1, 2, 3], 'value': {4, 5, 6}}",
            "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345": None,
            "urn:fz-juelich.de:alpaca:object:Python:test.OutputObject:54321": None,
        }

        for node, node_attrs in graph.graph.nodes(data=True):
            if node_attrs['type'] == 'object':
                expected_value = node_values_by_id[node]
                self.assertEqual(expected_value, node_attrs.get('value', None))


class GraphTimeIntervalTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        ttl_path = Path(__file__).parent / "res"
        cls.input_file = ttl_path / "parallel_graph.ttl"
        alpaca_setting('authority', "my-authority")

    def test_use_time_interval(self):
        graph = ProvenanceGraph(self.input_file, time_intervals=True)
        for _, attrs in graph.graph.nodes(data=True):
            self.assertFalse("gephi_interval" in attrs)
            self.assertTrue("Time Interval" in attrs)

    def test_not_use_time_interval(self):
        graph = ProvenanceGraph(self.input_file, time_intervals=False)
        for _, attrs in graph.graph.nodes(data=True):
            self.assertFalse("gephi_interval" in attrs)
            self.assertFalse("Time Interval" in attrs)

    def test_intervals(self):
        expected_intervals = {
"urn:fz-juelich.de:alpaca:object:Python:__main__.OutputObject:3934c99ea6197963f4bc7413932f6ce6dd800b08": "<[3.0,3.0]>",
            "urn:fz-juelich.de:alpaca:object:Python:__main__.OutputObject:4ef19b49bcf029faae5349020a54096d53398c95": "<[1.0,1.0]>",
            "urn:fz-juelich.de:alpaca:object:Python:__main__.OutputObject:93f4a32cb869a3e115e3382fd0fd49ab4ea0c8df": "<[2.0,2.0]>",
            "urn:fz-juelich.de:alpaca:object:Python:__main__.OutputObject:97ce94acf4ec4e2cb7d1319b798dbdd187df9558": "<[4.0,4.0]>",
            "urn:fz-juelich.de:alpaca:object:Python:builtins.list:f801594e5cebdc73ba8815e8ad66cab5cd86d2bf": "<[1.0,1.0];[2.0,2.0];[3.0,3.0];[4.0,4.0]>",
            "urn:fz-juelich.de:alpaca:function_execution:Python:4ff615bf10e589799a96729fdf19df67dc8b5fb03090a934107074b5c09b5393:13495a29-65e6-4853-90b1-05bb4dba9040:__main__.process#3dbe5e02-a5e6-48b6-8cb8-e3f0447d7a40": "<[1.0,1.0]>",
            "urn:fz-juelich.de:alpaca:function_execution:Python:4ff615bf10e589799a96729fdf19df67dc8b5fb03090a934107074b5c09b5393:13495a29-65e6-4853-90b1-05bb4dba9040:__main__.process#6ef55dd9-35f5-4519-aed5-80906c7fa341": "<[4.0,4.0]>",
                        "urn:fz-juelich.de:alpaca:function_execution:Python:4ff615bf10e589799a96729fdf19df67dc8b5fb03090a934107074b5c09b5393:13495a29-65e6-4853-90b1-05bb4dba9040:__main__.process#7e3565c0-4313-4229-a0dc-8fa81e4301a1": "<[3.0,3.0]>",
                        "urn:fz-juelich.de:alpaca:function_execution:Python:4ff615bf10e589799a96729fdf19df67dc8b5fb03090a934107074b5c09b5393:13495a29-65e6-4853-90b1-05bb4dba9040:__main__.process#f635dbb8-ad01-4c3d-99ca-5496940143cc": "<[2.0,2.0]>",
                        "urn:fz-juelich.de:alpaca:object:Python:__main__.InputObject:84fa33edca00abb3c664c3b994e455ae10fbefa1": "<[2.0,2.0]>",
            "urn:fz-juelich.de:alpaca:object:Python:__main__.InputObject:9dbee0f2b42ba928138d4eb3cc3059f2d7086716": "<[4.0,4.0]>",
            "urn:fz-juelich.de:alpaca:object:Python:__main__.InputObject:b443853aa145342288afaae4f68b6b421683f411": "<[1.0,1.0]>",
            "urn:fz-juelich.de:alpaca:object:Python:__main__.InputObject:eed23509f67bfc5dd108fe361ce57a1b9737a286": "<[3.0,3.0]>",
        }
        graph = ProvenanceGraph(self.input_file)
        for node, time_interval in graph.graph.nodes(data='Time Interval'):
            self.assertEqual(time_interval, expected_intervals[node])

    def test_aggregation_without_intervals(self):
        graph = ProvenanceGraph(self.input_file, time_intervals=False)
        aggregated = graph.aggregate({})
        for _, attrs in aggregated.nodes(data=True):
            self.assertFalse("gephi_interval" in attrs)
            self.assertFalse("Time Interval" in attrs)


class GraphAggregationTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ttl_path = Path(__file__).parent / "res"
        input_file = cls.ttl_path / "parallel_graph.ttl"
        cls.graph = ProvenanceGraph(input_file, attributes=['shape',
                                                            'metadata', 'id'])
        alpaca_setting('authority', "my-authority")

    def test_serialization(self):
        temp_dir = tempfile.TemporaryDirectory(dir=self.ttl_path, suffix="tmp")
        output_path = Path(temp_dir.name)

        gexf_file = output_path / "test.gexf"
        self.graph.aggregate({}, output_file=gexf_file)
        self.assertTrue(gexf_file.exists())

        graphml_file = output_path / "test.graphml"
        self.graph.aggregate({}, output_file=graphml_file)
        self.assertTrue(graphml_file.exists())

        with self.assertRaises(ValueError):
            self.graph.aggregate({}, output_file=output_path / "test.invalid")

    def test_overall_aggregation(self):
        aggregated = self.graph.aggregate({}, use_function_parameters=False,
                                          output_file=None)
        nodes = aggregated.nodes

        self.assertEqual(len(nodes), 4)

        expected_values_per_node = {
            'OutputObject': {'metadata': "0;1",
                             'shape': "(2,);(3,);(4,);(5,)"},
            'InputObject': {'metadata': "5",
                            'shape': "(2,);(3,);(4,);(5,)"},
            'process': {'process:value': "0;1;2;3"},
            'list': {}
        }

        all_labels = [nodes[node]['label'] for node in nodes]
        counts = Counter(all_labels)
        self.assertEqual(counts['OutputObject'], 1)
        self.assertEqual(counts['InputObject'], 1)
        self.assertEqual(counts['process'], 1)
        self.assertEqual(counts['list'], 1)

        for node, attrs in nodes.items():
            label = attrs['label']
            with self.subTest(f"Node label {label}"):
                self.assertTrue(label in expected_values_per_node)
                for key, value in expected_values_per_node[label].items():
                    self.assertEqual(attrs[key], value)

    def test_aggregation_by_callable(self):
        graph_file = self.ttl_path / "multiple_file_output.ttl"

        # Non-aggregated graph
        graph = ProvenanceGraph(graph_file)

        # Aggregate without attributes
        aggregated = graph.aggregate({}, output_file=None)

        # Aggregate separating by file path in File nodes
        aggregated_path = graph.aggregate({'File': ('File_path',)},
                                          output_file=None)

        # Aggregate using a callable to separate files which path starts with
        # "/outputs/"
        is_cut_plot = lambda g, n, d: d['File_path'].startswith("/outputs/")
        aggregated_callable = graph.aggregate({'File': (is_cut_plot,)},
                                              output_file=None)

        # Define a dictionary with the expected values for each case, that
        # are used in subtests below
        tests = {
            'non_aggregated': {'graph': graph.graph, 'length': 10,
                               'counts': {'InputObject': 3,
                                          'plot_function': 3,
                                          'cut_function': 1,
                                          'File': 3},
                               'paths': ["/full.png",
                                         "/outputs/1.png",
                                         "/outputs/2.png"]
                               },

            'aggregated': {'graph': aggregated, 'length': 5,
                           'counts': {'InputObject': 2,
                                      'plot_function': 1,
                                      'cut_function': 1,
                                      'File': 1},
                           'paths': "/full.png;/outputs/1.png;/outputs/2.png"
                           },

            'aggregated_path': {'graph': aggregated_path, 'length': 10,
                                'counts': {'InputObject': 3,
                                           'plot_function': 3,
                                           'cut_function': 1,
                                           'File': 3},
                                'paths': ["/full.png",
                                          "/outputs/1.png",
                                          "/outputs/2.png"]
                                },
            'aggregated_callable': {'graph': aggregated_callable, 'length': 7,
                                    'counts': {'InputObject': 2,
                                               'plot_function': 2,
                                               'cut_function': 1,
                                               'File': 2},
                                    'paths': ["/full.png",
                                              "/outputs/1.png;/outputs/2.png"]
                                    },
        }

        for key, expected in tests.items():
            with self.subTest(f"Graph {key}"):
                test_graph = expected['graph']
                nodes = test_graph.nodes
                self.assertEqual(len(nodes), expected['length'])

                # Check if node counts is as expected
                all_labels = [nodes[node]['label'] for node in nodes]
                counts = Counter(all_labels)
                for label, count in expected['counts'].items():
                    self.assertEqual(counts[label], count)

                # Check if file paths in the node are as expected
                paths = expected['paths']
                for node, attrs in nodes.items():
                    # Check value of file paths in File nodes
                    if attrs['label'] == "File":
                        if isinstance(paths, list):
                            self.assertTrue(attrs['File_path'] in paths)
                        else:
                            self.assertEqual(attrs['File_path'], paths)

    def test_aggregation_by_attribute_with_missing(self):
        aggregated = self.graph.aggregate({'InputObject': ('id',)},
                                          use_function_parameters=False,
                                          output_file=None)
        nodes = aggregated.nodes

        self.assertEqual(len(nodes), 5)

        expected_values_per_node = {
            'OutputObject': {'metadata': "0;1",
                             'shape': "(2,);(3,);(4,);(5,)"},
            'InputObject': {'metadata': "5",
                            'shape': ["(2,)", "(3,);(4,);(5,)"],
                            'id': ["1", None]},
            'process': {'process:value': "0;1;2;3"},
            'list': {}
        }

        all_labels = [nodes[node]['label'] for node in nodes]
        counts = Counter(all_labels)
        self.assertEqual(counts['OutputObject'], 1)
        self.assertEqual(counts['InputObject'], 2)
        self.assertEqual(counts['process'], 1)
        self.assertEqual(counts['list'], 1)

        for node, attrs in nodes.items():
            label = attrs['label']
            with self.subTest(f"Node label {label}"):
                self.assertTrue(label in expected_values_per_node)
                for key, value in expected_values_per_node[label].items():
                    attr_val = attrs[key] if key in attrs else None
                    if not isinstance(value, list):
                        self.assertEqual(attr_val, value)
                    else:
                        self.assertTrue(attr_val in value)

    def test_aggregation_by_attribute(self):
        aggregated = self.graph.aggregate({'InputObject': ('shape',)},
                                          use_function_parameters=False,
                                          output_file=None)
        nodes = aggregated.nodes

        self.assertEqual(len(nodes), 7)

        expected_values_per_node = {
            'OutputObject': {'metadata': "0;1",
                             'shape': "(2,);(3,);(4,);(5,)"},
            'InputObject': {'metadata': "5",
                            'shape': ["(2,)", "(3,)", "(4,)", "(5,)"]},
            'process': {'process:value': "0;1;2;3"},
            'list': {}
        }

        all_labels = [nodes[node]['label'] for node in nodes]
        counts = Counter(all_labels)
        self.assertEqual(counts['OutputObject'], 1)
        self.assertEqual(counts['InputObject'], 4)
        self.assertEqual(counts['process'], 1)
        self.assertEqual(counts['list'], 1)

        for node, attrs in nodes.items():
            label = attrs['label']
            with self.subTest(f"Node label {label}"):
                self.assertTrue(label in expected_values_per_node)
                for key, value in expected_values_per_node[label].items():
                    if not isinstance(value, list):
                        self.assertEqual(attrs[key], value)
                    else:
                        self.assertTrue(attrs[key] in value)

    def test_aggregation_by_attribute_with_function(self):
        aggregated = self.graph.aggregate({'InputObject': ('shape',)},
                                          use_function_parameters=True,
                                          output_file=None)
        nodes = aggregated.nodes

        self.assertEqual(len(nodes), 10)

        expected_values_per_node = {
            'OutputObject': {'metadata': "0;1",
                             'shape': "(2,);(3,);(4,);(5,)"},
            'InputObject': {'metadata': "5",
                            'shape': ["(2,)", "(3,)", "(4,)", "(5,)"]},
            'process': {'process:value': ["0", "1", "2", "3"]},
            'list': {}
        }

        all_labels = [nodes[node]['label'] for node in nodes]
        counts = Counter(all_labels)
        self.assertEqual(counts['OutputObject'], 1)
        self.assertEqual(counts['InputObject'], 4)
        self.assertEqual(counts['process'], 4)
        self.assertEqual(counts['list'], 1)

        for node, attrs in nodes.items():
            label = attrs['label']
            with self.subTest(f"Node label {label}"):
                self.assertTrue(label in expected_values_per_node)
                for key, value in expected_values_per_node[label].items():
                    if not isinstance(value, list):
                        self.assertEqual(attrs[key], value)
                    else:
                        self.assertTrue(attrs[key] in value)

    def test_aggregation_member_info(self):
        aggregated = self.graph.aggregate({}, use_function_parameters=False,
                                          output_file=None)
        nodes = aggregated.nodes

        self.assertEqual(len(nodes), 4)

        expected = {
            'OutputObject': {'member_count': 4,
                             'members': "urn:fz-juelich.de:alpaca:object:Python:__main__.OutputObject:4ef19b49bcf029faae5349020a54096d53398c95;urn:fz-juelich.de:alpaca:object:Python:__main__.OutputObject:3934c99ea6197963f4bc7413932f6ce6dd800b08;urn:fz-juelich.de:alpaca:object:Python:__main__.OutputObject:93f4a32cb869a3e115e3382fd0fd49ab4ea0c8df;urn:fz-juelich.de:alpaca:object:Python:__main__.OutputObject:97ce94acf4ec4e2cb7d1319b798dbdd187df9558"},
            'InputObject': {'member_count': 4,
                            'members': "urn:fz-juelich.de:alpaca:object:Python:__main__.InputObject:84fa33edca00abb3c664c3b994e455ae10fbefa1;urn:fz-juelich.de:alpaca:object:Python:__main__.InputObject:9dbee0f2b42ba928138d4eb3cc3059f2d7086716;urn:fz-juelich.de:alpaca:object:Python:__main__.InputObject:eed23509f67bfc5dd108fe361ce57a1b9737a286;urn:fz-juelich.de:alpaca:object:Python:__main__.InputObject:b443853aa145342288afaae4f68b6b421683f411"},
            'process': {'member_count': 4,
                        'members': "urn:fz-juelich.de:alpaca:function_execution:Python:4ff615bf10e589799a96729fdf19df67dc8b5fb03090a934107074b5c09b5393:13495a29-65e6-4853-90b1-05bb4dba9040:__main__.process#6ef55dd9-35f5-4519-aed5-80906c7fa341;urn:fz-juelich.de:alpaca:function_execution:Python:4ff615bf10e589799a96729fdf19df67dc8b5fb03090a934107074b5c09b5393:13495a29-65e6-4853-90b1-05bb4dba9040:__main__.process#7e3565c0-4313-4229-a0dc-8fa81e4301a1;urn:fz-juelich.de:alpaca:function_execution:Python:4ff615bf10e589799a96729fdf19df67dc8b5fb03090a934107074b5c09b5393:13495a29-65e6-4853-90b1-05bb4dba9040:__main__.process#3dbe5e02-a5e6-48b6-8cb8-e3f0447d7a40;urn:fz-juelich.de:alpaca:function_execution:Python:4ff615bf10e589799a96729fdf19df67dc8b5fb03090a934107074b5c09b5393:13495a29-65e6-4853-90b1-05bb4dba9040:__main__.process#f635dbb8-ad01-4c3d-99ca-5496940143cc"},
            'list': {'member_count': 1,
                     'members': "urn:fz-juelich.de:alpaca:object:Python:builtins.list:f801594e5cebdc73ba8815e8ad66cab5cd86d2bf"}
        }

        for node, attrs in nodes.items():
            label = attrs['label']
            with self.subTest(f"Node label {label}"):
                expected_info = expected[label]
                self.assertTrue('members' in attrs)
                self.assertEqual(attrs['member_count'],
                                 expected_info['member_count'])
                graph_members = sorted(attrs['members'].split(";"))
                expected_members = sorted(expected_info['members'].split(";"))
                self.assertListEqual(graph_members, expected_members)

    def test_aggregation_member_info_count_only(self):
        aggregated = self.graph.aggregate({}, use_function_parameters=False,
                                          output_file=None,
                                          record_members=False)
        nodes = aggregated.nodes

        self.assertEqual(len(nodes), 4)

        expected_counts = {
            'OutputObject': 4,
            'InputObject': 4,
            'process': 4,
            'list': 1
        }

        for node, attrs in nodes.items():
            label = attrs['label']
            with self.subTest(f"Node label {label}"):
                self.assertTrue('members' not in attrs)
                self.assertEqual(attrs['member_count'], expected_counts[label])


class GraphMultipleSourcesTestCase(unittest.TestCase):
    @staticmethod
    def _node_comparison(attr_G1, attr_G2):
        # Functions will have different execution IDs and orders
        for ignored_attr in ('execution_id', 'execution_order'):
            for attr in (attr_G1, attr_G2):
                if ignored_attr in attr:
                    attr.pop(ignored_attr)

        return attr_G1 == attr_G2

    @classmethod
    def setUpClass(cls):
        ttl_path = Path(__file__).parent / "res"
        cls.single_graph = ttl_path / "single_graph.ttl"
        cls.multiple_graphs = [ttl_path / f"multi_graph_{i}.ttl"
                               for i in range(2)]
        alpaca_setting('authority', "my-authority")

    def test_multiple_sources(self):
        single_graph = ProvenanceGraph(self.single_graph,
                                       attributes='all',
                                       annotations='all',
                                       time_intervals=False)
        multiple_graph = ProvenanceGraph(*self.multiple_graphs,
                                         attributes='all',
                                         annotations='all',
                                         time_intervals=False)

        single_nx_graph = single_graph.graph
        multiple_nx_graph = multiple_graph.graph

        self.assertEqual(len(single_nx_graph.nodes), 8)
        self.assertEqual(len(multiple_nx_graph.nodes), 8)

        # All objects should be present and with the same data
        for data_node, data in single_graph.graph.nodes(data=True):
            if data['type'] == 'object':
                self.assertTrue(data_node in multiple_nx_graph.nodes)

        # Graphs should be equal, except for the function executions IDs and
        # order
        self.assertTrue(nx.is_isomorphic(single_nx_graph, multiple_nx_graph,
                                         node_match=self._node_comparison))


if __name__ == "__main__":
    unittest.main()



"""
This module implements a class to store and fetch information from the source
code of the frame that activated provenance tracking.

The main purpose is to retrieve the full multiline statements that generated
the call to a tracked function.

The class defined in this module is not intended to be used directly by
the user, but is used internally by the `decorator.Provenance` decorator.
"""

import inspect
import ast
import numpy as np


class _SourceCode(object):
    """
    Stores the source code of the frame that activated provenance tracking,
    and provides methods for retrieving execution statements by line number.

    Parameters
    ----------
    frame : inspect.frame
        The frame of the scope where provenance tracking was activated,
        from which the source code will be fetched.

    Attributes
    ----------
    source_name : str
        Name of the function that has the source code from `frame`.
        If this is the main script, the value will be '<module>'.
    source_lineno : str
        Absolute line number in the script file where the code starts.
    ast_tree : ast.Node
        Parsed AST tree of the source code from `frame`.
    source_code_lines : np.ndarray
        Lines of the code from `frame`.
    """

    def __init__(self, frame):

        # Get the name of the function where activate was called
        self.source_name = inspect.getframeinfo(frame).function

        # Get current line in the frame
        exec_line = inspect.getlineno(frame)

        # Get the list with all the lines of the code being tracked
        code_lines = inspect.getsourcelines(frame)[0]

        # Clean any decorators (this happens when we are tracking inside a
        # function like `main`). The AST parser cannot deal with them.
        cur_line = 0
        while code_lines[cur_line].strip().startswith('@'):
            cur_line += 1
        parsed_lines = code_lines[cur_line:]

        # Parse and store the source code AST
        self.ast_tree = ast.parse("".join(parsed_lines).strip())

        # Set code start line. If the `provenance.activate` function was
        # called in the main script body, the name will be <module> and code
        # starts at line 1. If it was called inside a function (e.g. `main`),
        # we need to get the start line from the frame. In this case, the line
        # number is the position where the activate function was called
        if self.source_name == '<module>':
            self.source_lineno = 1
        else:
            activate_line = self._find_activate_line(self.ast_tree,
                                                     self.source_name)
            self.source_lineno = exec_line - (activate_line - 1)

        # Store the actual source lines. NumPy array is used to facilitate
        # slicing with masks
        self.source_code_lines = np.array(parsed_lines)

        # Build the mapping arrays, to fetch the statements later
        self._statement_lines, self._source_lines = \
            self._build_line_map(self.ast_tree, self.source_lineno,
                                 self.source_code_lines)

    @staticmethod
    def _find_activate_line(full_ast, function_name):
        # This function creates an AST and finds the location of the function
        # `activate` function call statement. We need this to find the correct
        # the position in the source code list

        if isinstance(full_ast.body[0], ast.FunctionDef):

            function_def = full_ast.body[0]

            if function_def.name == function_name:

                for statement in function_def.body:
                    for node in ast.walk(statement):
                        if ((isinstance(node, ast.Call)) and
                                (node.func.id in ('activate',))):
                            return statement.lineno
        return 0

    @staticmethod
    def _build_line_map(ast_tree, start_line_number, source_code_lines):
        # This function analyzes the AST structure of the code to fetch the
        # start and end lines of each statement. A mapping of each script line
        # to the actual code is also returned, that is used later when
        # fetching the full statements.

        # We extract a stack with all nodes in the script/function body. To
        # correct the starting line if provenance is tracked inside a function
        # (e.g., `def main():`), we set a flag to use later
        is_function = False
        if (len(ast_tree.body) == 1 and
                isinstance(ast_tree.body[0], ast.FunctionDef)):
            # We are tracking inside a function
            code_nodes = ast_tree.body[0].body
            is_function = True
        else:
            # We are tracking from the script root
            code_nodes = ast_tree.body

        # Build the list with line numbers of each main node in the
        # script/function body. These are stored in `statement_lines_numbers`
        # array, where column 0 is the starting line of the statement, and
        # column 1 the end line. The line information from the AST is relative
        # to the scope of code, i.e., for code inside a function, the first
        # line of the `def` statement is line 1. We correct this later after
        # having the full array.
        statement_lines_numbers = list()

        # We process node by node. Whenever code blocks are identified, all
        # nodes in its body are pushed to the `code_nodes` stack
        while code_nodes:
            node = code_nodes.pop(0)
            if hasattr(node, 'body'):
                # Another code block (e.g., if, for, while)
                # Just add the nodes in the body for further processing
                code_nodes.extend(node.body)

                # If `else` block is present, add it as well
                if hasattr(node, 'orelse') and node.orelse:
                    code_nodes.extend(node.orelse)

            else:
                # A statement. Find the maximum line number
                end_lines = [child.lineno for child in ast.walk(node) if
                             'lineno' in child._attributes]
                statement_lines_numbers.append((node.lineno, max(end_lines)))

        # Convert list to the final array, allowing easy masking
        statement_lines_numbers = sorted(statement_lines_numbers,
                                         key=lambda x: x[0])
        statement_lines_numbers = np.asarray(statement_lines_numbers)

        # Correct line numbers if source code is from function
        # In the end, `statement_lines_numbers` ranges are with respect to
        # the actual line numbers in the script, for each statement.

        if is_function:
            statement_lines_numbers += (start_line_number - 1)

        # Create an array with the line number of each line in
        # `source_code_lines` array. This will be used for masking
        source_lines_numbers = np.arange(
            start_line_number, start_line_number + source_code_lines.shape[0])

        return statement_lines_numbers, source_lines_numbers

    def extract_multiline_statement(self, line_number):
        """
        Fetch all code lines in case `line_number` contains a statement that
        is the end or part of a multiline statement.

        Parameters
        ----------
        line_number : int
            Line number from :attr:`source_code_lines`.

        Returns
        -------
        str or None
            The code corresponding to the full statement, or None if no
            statement was found in that line.
        """
        # Find the start and end line of the statement identified by
        # `line_number`
        line_diff = self._statement_lines[:, 0] - line_number
        nearest_mask = line_diff <= 0

        if not np.any(nearest_mask):
            return None

        nearest_number_index = np.argmax(line_diff[nearest_mask])
        statement_start, statement_end = \
            self._statement_lines[nearest_number_index, :]

        if line_number > statement_end:
            return None

        # Obtain the mask to get the source code between the start and end
        # lines
        line_mask = np.logical_and(self._source_lines >= statement_start,
                                   self._source_lines <= statement_end)

        # Retrieve the lines and join in a single string
        lines = self.source_code_lines[line_mask]
        statement = "".join(lines).strip()
        return statement



import unittest

import numpy as np

from alpaca.alpaca_types import File, DataObject
from alpaca.data_information import _FileInformation, _ObjectInformation

from pathlib import Path
import joblib
import uuid


class ObjectClass(object):
    """
    Class used to test hashing and getting data from custom objects
    """
    def __init__(self, param):
        self.param = param
        self.attribute = "an object class"


class FileInformationTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.file_path = Path(__file__).parent.absolute() / "res"

    def test_file_info_sha256(self):
        file_input = self.file_path / "file_input.txt"
        file_info = _FileInformation(file_input)
        info = file_info.info()
        self.assertIsInstance(info, File)
        self.assertEqual(info.hash_type, "sha256")
        self.assertEqual(info.hash, "96ccc1380e069667069acecea3e2ab559441657807e0a86d14f49028710ddb3a")
        self.assertEqual(info.path, file_input)

    def test_file_info_comparison(self):
        file_info_1 = _FileInformation(self.file_path / "file_input.txt")
        file_info_2 = _FileInformation(self.file_path / "file_input.txt")
        file_info_3 = _FileInformation(self.file_path / "file_output.txt")

        self.assertTrue(file_info_1 == file_info_2)

        with self.assertRaises(TypeError):
            comparison = file_info_1 == "other type"

        self.assertFalse(file_info_3 == file_info_1)

    def test_repr(self):
        file_info = _FileInformation(self.file_path / "file_input.txt")
        expected_str = "file_input.txt: [sha256] 96ccc1380e069667069acece" \
                       "a3e2ab559441657807e0a86d14f49028710ddb3a"
        self.assertEqual(str(file_info), expected_str)


class ObjectInformationTestCase(unittest.TestCase):

    def test_numpy_array(self):
        numpy_array_int = np.array([[1, 2, 3, 4],
                                    [5, 6, 7, 8],
                                    [9, 10, 11, 12]], dtype=np.int64)
        numpy_array_float = np.array([[1, 2, 3, 4],
                                      [5, 6, 7, 8],
                                      [9, 10, 11, 12]], dtype=np.float64)

        object_info = _ObjectInformation()
        info_int = object_info.info(numpy_array_int)
        info_float = object_info.info(numpy_array_float)

        self.assertIsInstance(info_int, DataObject)
        self.assertIsInstance(info_float, DataObject)

        self.assertEqual(info_int.type, "numpy.ndarray")
        self.assertEqual(info_float.type, "numpy.ndarray")
        self.assertEqual(info_int.details['shape'], (3, 4))
        self.assertEqual(info_float.details['shape'], (3, 4))
        self.assertEqual(info_int.details['dtype'], np.int64)
        self.assertEqual(info_float.details['dtype'], np.float64)
        self.assertEqual(info_int.id, id(numpy_array_int))
        self.assertEqual(info_float.id, id(numpy_array_float))
        self.assertEqual(info_int.hash_method, "joblib_SHA1")
        self.assertEqual(info_float.hash_method, "joblib_SHA1")
        self.assertEqual(info_int.hash, joblib.hash(numpy_array_int,
                                                    hash_name='sha1'))
        self.assertEqual(info_float.hash, joblib.hash(numpy_array_float,
                                                      hash_name='sha1'))

    def test_memoization(self):
        array = np.array([1, 2, 3])
        object_info = _ObjectInformation()
        array_id = id(array)

        self.assertDictEqual(object_info._hash_memoizer, {})
        self.assertFalse(array_id in object_info._hash_memoizer)
        info_pre = object_info.info(array)

        self.assertTrue(array_id in object_info._hash_memoizer)
        info_post = object_info.info(array)
        self.assertEqual(info_pre, info_post)

    def test_none(self):
        object_info = _ObjectInformation()
        info = object_info.info(None)
        self.assertIsInstance(info.hash, uuid.UUID)
        self.assertEqual(info.type, "builtins.NoneType")
        self.assertEqual(info.hash_method, "UUID")
        self.assertDictEqual(info.details, {})

    def test_store_value_requested(self):
        object_info = _ObjectInformation(store_values=['builtins.dict'])
        test_dict = dict(key=['3', '4'])
        info = object_info.info(test_dict)
        self.assertEqual(info.hash, joblib.hash(test_dict, hash_name='sha1'))
        self.assertEqual(info.type, "builtins.dict")
        self.assertEqual(info.hash_method, "joblib_SHA1")
        self.assertDictEqual(info.details, {})
        self.assertEqual(info.value, "{'key': ['3', '4']}")

    def test_store_value_not_requested(self):
        object_info = _ObjectInformation()
        test_dict = dict(key=['3', '4'])
        info = object_info.info(test_dict)
        self.assertEqual(info.hash, joblib.hash(test_dict, hash_name='sha1'))
        self.assertEqual(info.type, "builtins.dict")
        self.assertEqual(info.hash_method, "joblib_SHA1")
        self.assertDictEqual(info.details, {})
        self.assertEqual(info.value, None)

    def test_store_value_builtins(self):
        object_info = _ObjectInformation()
        info = object_info.info(5)
        self.assertEqual(info.hash, joblib.hash(5, hash_name='sha1'))
        self.assertEqual(info.type, "builtins.int")
        self.assertEqual(info.hash_method, "joblib_SHA1")
        self.assertDictEqual(info.details, {})
        self.assertEqual(info.value, 5)

    def test_custom_class(self):
        custom_object_1 = ObjectClass(param=4)
        custom_object_2 = ObjectClass(param=3)
        object_info = _ObjectInformation()

        info_1 = object_info.info(custom_object_1)
        self.assertEqual(info_1.details['param'], 4)
        self.assertEqual(info_1.details['attribute'], "an object class")
        self.assertEqual(info_1.id, id(custom_object_1))
        self.assertEqual(info_1.type, "test_data_information.ObjectClass")
        self.assertEqual(info_1.hash_method, "joblib_SHA1")
        self.assertEqual(info_1.hash, joblib.hash(custom_object_1,
                                                  hash_name='sha1'))

        info_2 = object_info.info(custom_object_2)
        self.assertEqual(info_2.details['param'], 3)
        self.assertEqual(info_2.details['attribute'], "an object class")
        self.assertEqual(info_2.id, id(custom_object_2))
        self.assertEqual(info_2.type, "test_data_information.ObjectClass")
        self.assertEqual(info_2.hash_method, "joblib_SHA1")
        self.assertEqual(info_2.hash, joblib.hash(custom_object_2,
                                                  hash_name='sha1'))

    def test_use_builtin_hash_simple(self):
        custom_object = ObjectClass(param=4)
        object_info = _ObjectInformation(
            use_builtin_hash=['test_data_information'])
        info = object_info.info(custom_object)
        self.assertEqual(info.details['param'], 4)
        self.assertEqual(info.details['attribute'], "an object class")
        self.assertEqual(info.id, id(custom_object))
        self.assertEqual(info.type, "test_data_information.ObjectClass")
        self.assertEqual(info.hash_method, "Python_hash")
        self.assertEqual(info.hash, hash(custom_object))

    def test_use_builtin_hash_container_list(self):
        custom_object_1 = ObjectClass(param=4)
        custom_object_2 = ObjectClass(param=3)
        object_info = _ObjectInformation(
            use_builtin_hash=['test_data_information'])

        container = [custom_object_1, custom_object_2]
        info = object_info.info(container)
        self.assertEqual(info.id, id(container))
        self.assertEqual(info.type, "builtins.list")
        self.assertEqual(info.hash_method, "Python_hash")

        expected_hashes = [hash(obj) for obj in container]

        self.assertEqual(info.hash, joblib.hash(tuple(expected_hashes),
                                                hash_name='sha1'))

    def test_use_builtin_hash_container_numpy_array(self):
        custom_object_1 = ObjectClass(param=4)
        custom_object_2 = ObjectClass(param=3)
        custom_object_3 = ObjectClass(param=7)
        custom_object_4 = ObjectClass(param=5)
        object_info = _ObjectInformation(
            use_builtin_hash=['test_data_information'])

        container = np.array([[custom_object_1, custom_object_2],
                              [custom_object_3, custom_object_4]])
        info = object_info.info(container)
        self.assertEqual(info.id, id(container))
        self.assertEqual(info.type, "numpy.ndarray")
        self.assertEqual(info.hash_method, "Python_hash")

        expected_hashes = []
        for row in container:
            for element in row:
                expected_hashes.append(hash(element))

        self.assertEqual(info.hash,
                         joblib.hash(tuple(expected_hashes), hash_name='sha1'))

    def test_use_builtin_hash_container_numpy_array_multiple_types(self):
        custom_object_1 = ObjectClass(param=4)
        custom_object_2 = ObjectClass(param=3)
        custom_object_3 = ObjectClass(param=7)
        object_info = _ObjectInformation(
            use_builtin_hash=['test_data_information'])

        container = np.array([[custom_object_1, custom_object_2],
                              [4, custom_object_3]])
        info = object_info.info(container)
        self.assertEqual(info.id, id(container))
        self.assertEqual(info.type, "numpy.ndarray")
        self.assertEqual(info.hash_method, "joblib_SHA1")

        self.assertEqual(info.hash, joblib.hash(container, hash_name='sha1'))


if __name__ == "__main__":
    unittest.main()



"""
This module contains utility functions to work with file names and extensions.

.. autofunction:: alpaca.utils.get_file_name

"""

from pathlib import Path


RDF_FILE_FORMAT_MAP = {
    'ttl': 'turtle',
    'rdf': 'xml',
    'json': 'json-ld',
}


def get_file_name(source, output_dir=None, extension=None, suffix=None):
    """
    Function that generates a file name with extension `extension` and the
    same base name as in `source`. The full path is based on `output_dir` if
    specified. Otherwise, it will be the same path as `source`. User and
    relative paths are expanded.

    Parameters
    ----------
    source : str or Path-like
        Source path or file name to generate the new file name. The base name
        will be considered.
    output_dir : str, optional
        If not None, the generated file name will have this path.
        Default: None
    extension : str, optional
        If not None, the extension of the generated file name will be changed
        to `extension`. If None, the same extension as `source` will be used.
        The extension may start with period.
        Default: None
    suffix : str, optional
        If not None, this will be added as a suffix to the base name of
        `source`, before the extension.
        Default: None

    Returns
    -------
    str
        File name, according to the parameters selected. If both `output_dir`
        and `extension` are None, the result will be equal to `source`. The
        result path is absolute, with user and relative paths expanded.
    """
    if not isinstance(source, Path):
        source = Path(source)

    if suffix is not None:
        parent, name, ext = source.parent, source.stem, source.suffix
        name = f"{name}{suffix}"
        source = (parent / name).with_suffix(ext)

    if extension is not None:
        if not extension.startswith("."):
            extension = f".{extension}"
        base_name = source.with_suffix(extension)
    else:
        base_name = source

    if output_dir is not None:
        base_name = Path(output_dir) / base_name.name
    return str(base_name.expanduser().resolve().absolute())


def _get_file_format(file_name):
    # Returns a string describing the file format based on the extension in
    # `file_name`. Returns None if no extension.
    file_location = Path(file_name)

    extension = file_location.suffix
    if not extension.startswith('.'):
        return None
    file_format = extension[1:]

    return file_format


def _get_prov_file_format(file_name):
    # Returns a string describing the file format based on the extension in
    # `file_name`. `.rdf` files are described as XML, `.ttl` files are
    # described as Turtle, and `.json` are described as JSON-LD. Other
    # extensions are returned as provided. The return value is compatible with
    # RDFLib serialization format strings. Returns None if no extension.

    file_format = _get_file_format(file_name)

    if file_format and file_format in RDF_FILE_FORMAT_MAP:
        return RDF_FILE_FORMAT_MAP[file_format]

    return file_format



import unittest

import joblib
import datetime
import sys
from io import StringIO
import tempfile
from pathlib import Path

import rdflib
from rdflib.compare import graph_diff

import numpy as np
import quantities as pq
import neo

from alpaca import (Provenance, activate, deactivate, save_provenance,
                    print_history, alpaca_setting)
from alpaca.alpaca_types import (FunctionInfo, Container, DataObject, File)

# Define some data and expected values test tracking

TEST_ARRAY = np.array([1, 2, 3])
TEST_ARRAY_INFO = DataObject(hash=joblib.hash(TEST_ARRAY, hash_name='sha1'),
                             hash_method="joblib_SHA1",
                             type="numpy.ndarray", id=id(TEST_ARRAY),
                             details={'shape': (3,), 'dtype': np.int64},
                             value=None)

TEST_ARRAY_2 = np.array([4, 5, 6])
TEST_ARRAY_2_INFO = DataObject(hash=joblib.hash(TEST_ARRAY_2,
                                                hash_name='sha1'),
                               hash_method="joblib_SHA1",
                               type="numpy.ndarray", id=id(TEST_ARRAY_2),
                               details={'shape': (3,), 'dtype': np.int64},
                               value=None)

CONTAINER = [TEST_ARRAY, TEST_ARRAY_2]


# Define some functions to test tracking in different scenarios

@Provenance(inputs=['array'])
def simple_function(array, param1, param2):
    """ Takes a single input and outputs a single element"""
    return array + 3


@Provenance(inputs=['array'])
def simple_function_default(array, param1, param2=10):
    """ Takes a single input and outputs a single element.
    One kwarg is default
    """
    return array + 5


@Provenance(inputs=None, container_input=['arrays'])
def container_input_function(arrays, param1, param2):
    """ Takes a container input (e.g. list) and outputs a single element"""
    return np.mean(arrays)


@Provenance(inputs=['arrays'])
def varargs_function(*arrays, param1, param2):
    """ Takes a variable argument input and outputs a single element"""
    return np.mean(arrays)


@Provenance(inputs=['array'])
def multiple_outputs_function(array, param1, param2):
    """ Takes a single input and outputs multiple elements as a tuple"""
    return array + 3, array + 4


@Provenance(inputs=['array_1', 'array_2'])
def multiple_inputs_function(array_1, array_2, param1, param2):
    """ Takes multiple inputs and outputs a single element"""
    return array_1 + array_2


@Provenance(inputs=['array'], container_output=True)
def container_output_function(array, param1, param2):
    """
    Takes a single input and outputs multiple elements in a container.
    This function will have as tracked outputs all elements inside the first
    level, i.e., the two NumPy arrays.
    """
    return [array + i for i in range(3, 5)]


@Provenance(inputs=['array'], container_output=0)
def container_output_function_level_0(array, param1, param2):
    """
    Takes a single input and outputs multiple elements in a container.
    This function will have as tracked output the return list.
    Additional nodes from the list to each element inside the first level
    (i.e., the two NumPy arrays) will be added.
    """
    return [array + i for i in range(7, 9)]


@Provenance(inputs=['array'], container_output=1)
def container_output_function_level_1(array, param1, param2):
    """
    Takes a single input and outputs multiple elements in a container.
    This function will have as tracked output the return list.
    Additional nodes from the list to each element inside the first level
    (i.e., the two NumPy arrays) and from each array to all its elements
    (i.e., the integers in the second level) will be added.
    """
    return [array + i for i in range(2, 4)]


@Provenance(inputs=['array'], container_output=(0, 0))
def container_output_function_level_range_0_0(array, param1, param2):
    """
    Takes a single input and outputs multiple elements in a container.
    As we are requesting to track from output level zero to zero, this
    function will have as tracked output the return list. Additional nodes
    from the list to each element inside the first level
    (i.e., the two NumPy arrays) will be added.
    """
    return [array + i for i in range(1, 3)]


@Provenance(inputs=['array'], container_output=(0, 1))
def container_output_function_level_range_0_1(array, param1, param2):
    """
    Takes a single input and outputs multiple elements in a container.
    As we are requesting to track from output level zero to one, this function
    will have as tracked output the return list. Additional nodes from the
    list to each element inside the first level (i.e., the two NumPy arrays)
    and from each array to all its elements (i.e., the integers in the second
    level) will be added.
    """
    return [array + i for i in range(5, 7)]


@Provenance(inputs=['array'], container_output=(1, 1))
def container_output_function_level_range_1_1(array, param1, param2):
    """
    Takes a single input and outputs multiple elements in a container.
    As we are requesting to track from output level one to one, this function
    will have as tracked outputs all elements inside the first
    level, i.e., the two NumPy arrays. Additional nodes each array to all
    its elements (i.e., the integers in the second level) will be added.
    This option skips adding an output node for the first level, i.e., the
    list that contains the two arrays.
    """
    return [array + i for i in range(4, 6)]


@Provenance(inputs=['array'], container_output=True)
def dict_output_function(array, param1, param2):
    """ Takes as single input and outputs multiple elements in a dictionary """
    return {f"key.{i}": array + i + 3 for i in range(0, 2)}


@Provenance(inputs=['array'], container_output=1)
def dict_output_function_level(array, param1, param2):
    """ Takes as single input and outputs multiple elements in a dictionary """
    return {f"key.{i}": array + i + 3 for i in range(0, 2)}


class NonIterableContainer(object):

    def __init__(self, start):
        self.data = np.arange(start, start+3)

    def __getitem__(self, item):
        return  self.data[item]


@Provenance(inputs=[], container_output=0)
def non_iterable_container_output(param1):
    return NonIterableContainer(param1)


@Provenance(inputs=[])
def comprehension_function(param):
    return np.float64(param)


class NonIterableContainerOutputObject(object):
    def __init__(self, start):
        self._data = np.arange(start+1, start+4)

    def __getitem__(self, item):
        return  self._data[item]


NonIterableContainerOutputObject.__init__ = \
    Provenance(inputs=[], container_output=0)(NonIterableContainerOutputObject.__init__)


# Function to help verifying FunctionExecution tuples
def _check_function_execution(actual, exp_function, exp_input, exp_params,
                              exp_output, exp_arg_map, exp_kwarg_map,
                              exp_code_stmnt, exp_return_targets, exp_order,
                              test_case):
    # Check function
    test_case.assertTupleEqual(actual.function, exp_function)

    # Check inputs
    for input_arg, value in exp_input.items():
        test_case.assertTrue(input_arg in actual.input)
        actual_input = actual.input[input_arg]
        for attr in value._fields:
            actual_value = getattr(actual_input, attr)
            exp_value = getattr(value, attr)
            if attr != 'id' or (attr == 'id' and exp_value is not None):
                test_case.assertEqual(actual_value, exp_value)

    # Check parameters
    test_case.assertDictEqual(actual.params, exp_params)

    # Check outputs
    for output, value in exp_output.items():
        test_case.assertTrue(output in actual.output)
        actual_output = actual.output[output]
        for attr in value._fields:
            actual_value = getattr(actual_output, attr)
            exp_value = getattr(value, attr)
            if attr != 'id' or (attr == 'id' and exp_value is not None):
                test_case.assertEqual(actual_value, exp_value)

    # Check args and kwargs
    if actual.arg_map is not None:
        test_case.assertListEqual(actual.arg_map, exp_arg_map)
    else:
        test_case.assertIsNone(exp_arg_map)

    if actual.kwarg_map is not None:
        test_case.assertListEqual(actual.kwarg_map, exp_kwarg_map)
    else:
        test_case.assertIsNone(exp_kwarg_map)

    # Check other information
    test_case.assertEqual(actual.code_statement, exp_code_stmnt)
    test_case.assertListEqual(actual.return_targets, exp_return_targets)
    test_case.assertEqual(actual.order, exp_order)
    test_case.assertNotEqual(actual.execution_id, "")

    # Check time stamps are valid ISO dates
    test_case.assertIsInstance(
        datetime.datetime.fromisoformat(actual.time_stamp_start),
        datetime.datetime)
    test_case.assertIsInstance(
        datetime.datetime.fromisoformat(actual.time_stamp_end),
        datetime.datetime)


class ProvenanceDecoratorInterfaceFunctionsTestCase(unittest.TestCase):

    def test_activate_deactivate(self):
        activate(clear=True)
        simple_function(TEST_ARRAY, 1, 2)
        simple_function(TEST_ARRAY, 3, 4)
        deactivate()
        simple_function(TEST_ARRAY, 5, 6)

        self.assertEqual(len(Provenance.history), 2)
        self.assertEqual(Provenance.history[0].code_statement,
                         "simple_function(TEST_ARRAY, 1, 2)")
        self.assertEqual(Provenance.history[1].code_statement,
                         "simple_function(TEST_ARRAY, 3, 4)")

    def test_save_provenance_show_progress(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, 2)
        deactivate()

        # Capture STDERR and serialize
        captured = StringIO()
        sys.stderr = captured
        save_provenance(file_name=None, show_progress=True)
        sys.stderr = sys.__stderr__

        captured_stderr = captured.getvalue()

        self.assertTrue("Serializing provenance history: 100%" in
                        captured_stderr)

    def test_save_provenance_no_progress(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, 2)
        deactivate()

        # Capture STDERR and serialize
        captured = StringIO()
        sys.stderr = captured
        save_provenance(file_name=None)
        sys.stderr = sys.__stderr__

        captured_stderr = captured.getvalue()

        self.assertEqual(captured_stderr, "")

    def test_save_provenance(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, 2)
        deactivate()

        # For every supported format, serialize to a string
        for output_format in ('json-ld', 'n3', 'nt', 'hext', 'pretty-xml',
                              'trig', 'turtle', 'longturtle', 'xml'):
            with self.subTest(f"Serialization format",
                              output_format=output_format):
                serialization = save_provenance(file_name=None,
                                                file_format=output_format)
                self.assertNotEqual(serialization, "")

        # For shortucts, test the expected serializations
        for short_output_format, output_format in (('ttl', 'turtle'),
                                                   ('rdf', 'xml'),
                                                   ('json', 'json-ld')):
            with self.subTest(f"Short serialization format",
                              short_format=short_output_format):
                short = save_provenance(None, file_format=short_output_format)
                serialization = save_provenance(None, file_format=output_format)

                short_graph = rdflib.Graph()
                short_graph.parse(StringIO(short), format=output_format)

                serialization_graph = rdflib.Graph()
                serialization_graph.parse(StringIO(serialization),
                                          format=output_format)

                self.assertTrue(short_graph.isomorphic(serialization_graph))
                _, in_first, in_second = graph_diff(short_graph,
                                                    serialization_graph)
                self.assertEqual(len(in_first), 0)
                self.assertEqual(len(in_second), 0)

    def test_save_provenance_no_capture(self):
        Provenance.clear()
        res = simple_function(TEST_ARRAY, 1, 2)

        serialization = save_provenance(None, file_format='turtle')
        self.assertIsNone(serialization)

    def test_print_history(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, 2)
        deactivate()

        expected_str = str(Provenance.history)

        # Capture STDOUT and print
        captured = StringIO()
        sys.stdout = captured
        print_history()
        sys.stdout = sys.__stdout__

        self.assertEqual(captured.getvalue().replace("\n", ""), expected_str)


class ProvenanceDecoratorFunctionsTestCase(unittest.TestCase):

    def test_get_module_version(self):
        expected_numpy_version = np.__version__

        numpy_version = Provenance._get_module_version("numpy")
        self.assertEqual(numpy_version, expected_numpy_version)

        numpy_version_submodule = Provenance._get_module_version(
            "numpy.random")
        self.assertEqual(numpy_version_submodule, expected_numpy_version)

        main_version = Provenance._get_module_version("__main__")
        self.assertEqual(main_version, "")

        invalid = Provenance._get_module_version("non_existent")
        self.assertEqual(invalid, "")

        none = Provenance._get_module_version(None)
        self.assertEqual(none, "")


class ProvenanceDecoratorInputOutputCombinationsTestCase(unittest.TestCase):

    def test_simple_function(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, 2)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 1, 'param2': 2},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = simple_function(TEST_ARRAY, 1, 2)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_simple_function_no_target(self):
        activate(clear=True)
        simple_function(TEST_ARRAY, param2=1, param1=2)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        # In this test we cannot know the id of the output, as it is not
        # stored in any variable. Let's get it from the history so that the
        # test does not fail
        output_id = Provenance.history[0].output[0].id

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=output_id,
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 2, 'param2': 1},
            exp_output={0: expected_output},
            exp_arg_map=['array'],
            exp_kwarg_map=['param1', 'param2'],
            exp_code_stmnt="simple_function(TEST_ARRAY, param2=1, param1=2)",
            exp_return_targets=[],
            exp_order=1,
            test_case=self)

    def test_kwargs_params(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, param2=2)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 1, 'param2': 2},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1'],
            exp_kwarg_map=['param2'],
            exp_code_stmnt="res = simple_function(TEST_ARRAY, 1, param2=2)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_kwargs_params_default(self):
        activate(clear=True)
        res = simple_function_default(TEST_ARRAY, 1)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+5, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function_default',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 1, 'param2': 10},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1'],
            exp_kwarg_map=['param2'],
            exp_code_stmnt="res = simple_function_default(TEST_ARRAY, 1)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_kwargs_params_default_override(self):
        activate(clear=True)
        res = simple_function_default(TEST_ARRAY, 1, 8)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+5, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function_default',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 1, 'param2': 8},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = simple_function_default(TEST_ARRAY, 1, 8)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_container_input_function(self):
        activate(clear=True)
        avg = container_input_function(CONTAINER, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)
        self.assertEqual(avg, 3.5)

        expected_output = DataObject(
            hash=joblib.hash(np.float64(3.5), hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.float64", id=id(avg),
            details={'shape': (), 'dtype': np.float64}, value=3.5)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('container_input_function',
                                      'test_decorator', ''),
            exp_input={'arrays': Container(tuple(
                [TEST_ARRAY_INFO, TEST_ARRAY_2_INFO]))},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['arrays', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="avg = container_input_function(CONTAINER, 3, 6)",
            exp_return_targets=['avg'],
            exp_order=1,
            test_case=self)

    def test_varargs_input_function(self):
        activate(clear=True)
        avg = varargs_function(*CONTAINER, param1=1, param2=2)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)
        self.assertEqual(avg, 3.5)

        expected_output = DataObject(
            hash=joblib.hash(np.float64(3.5), hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.float64", id=id(avg),
            details={'shape': (), 'dtype': np.float64}, value=3.5)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('varargs_function',
                                      'test_decorator', ''),
            exp_input={'arrays': Container(tuple(
                [TEST_ARRAY_INFO, TEST_ARRAY_2_INFO]))},
            exp_params={'param1': 1, 'param2': 2},
            exp_output={0: expected_output},
            exp_arg_map=['arrays'],
            exp_kwarg_map=['param1', 'param2'],
            exp_code_stmnt="avg = varargs_function(*CONTAINER, param1=1, param2=2)",
            exp_return_targets=['avg'],
            exp_order=1,
            test_case=self)

    def test_multiple_inputs_function(self):
        activate(clear=True)
        res = multiple_inputs_function(TEST_ARRAY, TEST_ARRAY_2, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+TEST_ARRAY_2, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('multiple_inputs_function',
                                      'test_decorator', ''),
            exp_input={'array_1': TEST_ARRAY_INFO,
                       'array_2': TEST_ARRAY_2_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['array_1', 'array_2', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = multiple_inputs_function(TEST_ARRAY, TEST_ARRAY_2, 3, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_multiple_outputs_function_elements(self):
        activate(clear=True)
        res1, res2 = multiple_outputs_function(TEST_ARRAY, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY+3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res1),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        expected_output_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY+4, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res2),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('multiple_outputs_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output_1, 1: expected_output_2},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res1, res2 = multiple_outputs_function(TEST_ARRAY, 3, 6)",
            exp_return_targets=['res1', 'res2'],
            exp_order=1,
            test_case=self)

    def test_multiple_outputs_function_tuple(self):
        activate(clear=True)
        res = multiple_outputs_function(TEST_ARRAY, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash((TEST_ARRAY+3, TEST_ARRAY+4), hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="builtins.tuple", id=id(res),
            details={}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('multiple_outputs_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = multiple_outputs_function(TEST_ARRAY, 3, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_container_output_function(self):
        activate(clear=True)
        res = container_output_function(TEST_ARRAY, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[0]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        expected_output_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 4, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[1]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('container_output_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output_1, 1: expected_output_2},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = container_output_function(TEST_ARRAY, 3, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_container_output_function_level_0(self):
        activate(clear=True)
        res = container_output_function_level_0(TEST_ARRAY, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 3)

        expected_output = DataObject(
            hash=joblib.hash(res, hash_name="sha1"), hash_method="joblib_SHA1",
            type="builtins.list", id=id(res), details={}, value=None)

        expected_container_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 7, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[0]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        expected_container_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 8, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[1]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        # Check the subscript of each array with respect to the list returned
        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 0},
            exp_output={0: expected_container_1},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 1},
            exp_output={0: expected_container_2},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        # Main function execution
        _check_function_execution(
            actual=Provenance.history[2],
            exp_function=FunctionInfo('container_output_function_level_0',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = container_output_function_level_0(TEST_ARRAY, 3, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_container_output_function_level_1(self):
        activate(clear=True)
        res = container_output_function_level_1(TEST_ARRAY, 4, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 9)

        elements = [[], []]
        for idx, container in enumerate(res):
            for element in container:
                element_info = DataObject(
                    hash=joblib.hash(element, hash_name="sha1"),
                    hash_method="joblib_SHA1",
                    type="numpy.int64", id=None,
                    details={'shape': (), 'dtype': np.int64},
                    value=element)
                elements[idx].append(element_info)

        expected_output = DataObject(
            hash=joblib.hash(res, hash_name="sha1"), hash_method="joblib_SHA1",
            type="builtins.list", id=id(res), details={}, value=None)

        expected_container_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 2, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[0]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        expected_container_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[1]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        # Check subscript of each element with respect to the array
        containers = [expected_container_1, expected_container_2]
        for history_index, element_index in zip(
                (1, 2, 3, 5, 6, 7),
                ((0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2))):
            container = element_index[0]
            element = element_index[1]
            _check_function_execution(
                actual=Provenance.history[history_index],
                exp_function=FunctionInfo('subscript', '', ''),
                exp_input={0: containers[container]},
                exp_params={'index': element},
                exp_output={0: elements[container][element]},
                exp_arg_map=None,
                exp_kwarg_map=None,
                exp_code_stmnt=None,
                exp_return_targets=[],
                exp_order=None,
                test_case=self)

        # Check the subscript of each array with respect to the list returned
        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 0},
            exp_output={0: expected_container_1},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[4],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 1},
            exp_output={0: expected_container_2},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        # Main function execution
        _check_function_execution(
            actual=Provenance.history[8],
            exp_function=FunctionInfo('container_output_function_level_1',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 4, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = container_output_function_level_1(TEST_ARRAY, 4, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_container_output_function_level_range_0_0(self):
        # Should be similar to `container_output=0`
        activate(clear=True)
        res = container_output_function_level_range_0_0(TEST_ARRAY, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 3)

        expected_output = DataObject(
            hash=joblib.hash(res, hash_name="sha1"), hash_method="joblib_SHA1",
            type="builtins.list", id=id(res), details={}, value=None)

        expected_container_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 1, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[0]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        expected_container_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 2, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[1]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        # Check the subscript of each array with respect to the list returned
        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 0},
            exp_output={0: expected_container_1},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 1},
            exp_output={0: expected_container_2},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        # Main function execution
        _check_function_execution(
            actual=Provenance.history[2],
            exp_function=FunctionInfo('container_output_function_level_range_0_0',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = container_output_function_level_range_0_0(TEST_ARRAY, 3, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_container_output_function_level_range_0_1(self):
        # Should be similar to `container_output=1`
        activate(clear=True)
        res = container_output_function_level_range_0_1(TEST_ARRAY, 4, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 9)

        elements = [[], []]
        for idx, container in enumerate(res):
            for el_idx, element in enumerate(container):
                element_info = DataObject(
                    hash=joblib.hash(element, hash_name="sha1"),
                    hash_method="joblib_SHA1",
                    type="numpy.int64", id=None,
                    details={'shape': (), 'dtype': np.int64}, value=element)
                elements[idx].append(element_info)

        expected_output = DataObject(
            hash=joblib.hash(res, hash_name="sha1"), hash_method="joblib_SHA1",
            type="builtins.list", id=id(res), details={}, value=None)

        expected_container_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 5, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[0]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        expected_container_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 6, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[1]),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        # Check subscript of each element with respect to the array
        containers = [expected_container_1, expected_container_2]
        for history_index, element_index in zip(
                (1, 2, 3, 5, 6, 7),
                ((0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2))):
            container = element_index[0]
            element = element_index[1]
            _check_function_execution(
                actual=Provenance.history[history_index],
                exp_function=FunctionInfo('subscript', '', ''),
                exp_input={0: containers[container]},
                exp_params={'index': element},
                exp_output={0: elements[container][element]},
                exp_arg_map=None,
                exp_kwarg_map=None,
                exp_code_stmnt=None,
                exp_return_targets=[],
                exp_order=None,
                test_case=self)

        # Check the subscript of each array with respect to the list returned
        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 0},
            exp_output={0: expected_container_1},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[4],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 1},
            exp_output={0: expected_container_2},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        # Main function execution
        _check_function_execution(
            actual=Provenance.history[8],
            exp_function=FunctionInfo('container_output_function_level_range_0_1',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 4, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = container_output_function_level_range_0_1(TEST_ARRAY, 4, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_container_output_function_level_range_1_1(self):
        activate(clear=True)
        res = container_output_function_level_range_1_1(TEST_ARRAY, 4, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 7)

        elements = [[], []]
        for idx, container in enumerate(res):
            for el_idx, element in enumerate(container):
                element_info = DataObject(
                    hash=joblib.hash(element, hash_name="sha1"),
                    hash_method="joblib_SHA1",
                    type="numpy.int64", id=None,
                    details={'shape': (), 'dtype': np.int64},
                    value=element)
                elements[idx].append(element_info)

        expected_container_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 4, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[0]),
            details={'shape': (3,), 'dtype': np.int64},
            value=None)

        expected_container_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 5, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[1]),
            details={'shape': (3,), 'dtype': np.int64},
            value=None)

        # Check subscript of each element with respect to the array
        containers = [expected_container_1, expected_container_2]
        for history_index, element_index in zip(
                (0, 1, 2, 3, 4, 5),
                ((0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2))):
            container = element_index[0]
            element = element_index[1]
            _check_function_execution(
                actual=Provenance.history[history_index],
                exp_function=FunctionInfo('subscript', '', ''),
                exp_input={0: containers[container]},
                exp_params={'index': element},
                exp_output={0: elements[container][element]},
                exp_arg_map=None,
                exp_kwarg_map=None,
                exp_code_stmnt=None,
                exp_return_targets=[],
                exp_order=None,
                test_case=self)

        # Main function execution
        # There is no single list return directly from the function
        _check_function_execution(
            actual=Provenance.history[6],
            exp_function=FunctionInfo('container_output_function_level_range_1_1',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 4, 'param2': 6},
            exp_output={0: expected_container_1, 1: expected_container_2},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = container_output_function_level_range_1_1(TEST_ARRAY, 4, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_dict_output_function(self):
        activate(clear=True)
        res = dict_output_function(TEST_ARRAY, 3, 7)
        deactivate()

        self.assertEqual(len(Provenance.history), 3)

        expected_output = DataObject(
            hash=joblib.hash(res, hash_name="sha1"), hash_method="joblib_SHA1",
            type="builtins.dict", id=id(res), details={}, value=None)

        expected_container_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res['key.0']),
            details={'shape': (3,), 'dtype': np.int64},
            value=None)

        expected_container_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 4, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res['key.1']),
            details={'shape': (3,), 'dtype': np.int64},
            value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 'key.0'},
            exp_output={0: expected_container_1},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 'key.1'},
            exp_output={0: expected_container_2},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        # Main function execution
        _check_function_execution(
            actual=Provenance.history[2],
            exp_function=FunctionInfo('dict_output_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 7},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = dict_output_function(TEST_ARRAY, 3, 7)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_dict_output_function_level(self):
        activate(clear=True)
        res = dict_output_function_level(TEST_ARRAY, 3, 8)
        deactivate()

        self.assertEqual(len(Provenance.history), 9)

        elements = {'key.0': [], 'key.1': []}
        for key, container in res.items():
            for element in container:
                element_info = DataObject(
                    hash=joblib.hash(element, hash_name="sha1"),
                    hash_method="joblib_SHA1",
                    type="numpy.int64", id=None,
                    details={'shape': (), 'dtype': np.int64},
                    value=element)
                elements[key].append(element_info)

        expected_output = DataObject(
            hash=joblib.hash(res, hash_name="sha1"), hash_method="joblib_SHA1",
            type="builtins.dict", id=id(res), details={}, value=None)

        expected_container_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res['key.0']),
            details={'shape': (3,), 'dtype': np.int64},
            value=None)

        expected_container_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 4, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res['key.1']),
            details={'shape': (3,), 'dtype': np.int64},
            value=None)

        # Check subscript of each element with respect to the array
        containers = {
            'key.0': expected_container_1,
            'key.1': expected_container_2
        }
        for history_index, element_index in zip(
                (1, 2, 3, 5, 6, 7),
                (('key.0', 0), ('key.0', 1), ('key.0', 2),
                 ('key.1', 0), ('key.1', 1), ('key.1', 2))):
            container = element_index[0]
            element = element_index[1]
            _check_function_execution(
                actual=Provenance.history[history_index],
                exp_function=FunctionInfo('subscript', '', ''),
                exp_input={0: containers[container]},
                exp_params={'index': element},
                exp_output={0: elements[container][element]},
                exp_arg_map=None,
                exp_kwarg_map=None,
                exp_code_stmnt=None,
                exp_return_targets=[],
                exp_order=None,
                test_case=self)

        # Check subscript of each array with respect to the dictionary returned
        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 'key.0'},
            exp_output={0: expected_container_1},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[4],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: expected_output},
            exp_params={'index': 'key.1'},
            exp_output={0: expected_container_2},
            exp_arg_map=None,
            exp_kwarg_map=None,
            exp_code_stmnt=None,
            exp_return_targets=[],
            exp_order=None,
            test_case=self)

        # Main function execution
        _check_function_execution(
            actual=Provenance.history[8],
            exp_function=FunctionInfo('dict_output_function_level',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 8},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = dict_output_function_level(TEST_ARRAY, 3, 8)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_non_iterable_container_output(self):
        activate(clear=True)
        res = non_iterable_container_output(3)
        deactivate()

        self.assertEqual(len(Provenance.history), 4)

        elements = []
        for element in res:
            element_info = DataObject(
                hash=joblib.hash(element, hash_name="sha1"),
                hash_method="joblib_SHA1",
                type="numpy.int64", id=None,
                details={'shape': (), 'dtype': np.int64},
                value=element)
            elements.append(element_info)

        expected_output = DataObject(
            hash=joblib.hash(res, hash_name="sha1"), hash_method="joblib_SHA1",
            type="test_decorator.NonIterableContainer", id=id(res),
            details={'data': res.data}, value=None)

        # Check subscript of each element with respect to the container
        for history_index in (0, 1, 2):
            element = elements[history_index]
            _check_function_execution(
                actual=Provenance.history[history_index],
                exp_function=FunctionInfo('subscript', '', ''),
                exp_input={0: expected_output},
                exp_params={'index': history_index},
                exp_output={0: element},
                exp_arg_map=None,
                exp_kwarg_map=None,
                exp_code_stmnt=None,
                exp_return_targets=[],
                exp_order=None,
                test_case=self)

        # Main function execution
        _check_function_execution(
            actual=Provenance.history[3],
            exp_function=FunctionInfo('non_iterable_container_output',
                                      'test_decorator', ''),
            exp_input={},
            exp_params={'param1': 3},
            exp_output={0: expected_output},
            exp_arg_map=['param1'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = non_iterable_container_output(3)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_comprehensions(self):
        activate(clear=True)
        num_list = [comprehension_function(i) for i in range(3)]
        num_set = {comprehension_function(i) for i in range(3, 6)}
        num_dict = {i: comprehension_function(v) for i, v in enumerate(range(6, 9), start=1)}
        deactivate()

        self.assertEqual(len(Provenance.history), 9)
        self.assertEqual(len(num_list), 3)
        self.assertEqual(len(num_set), 3)
        self.assertEqual(list(num_dict.keys()), [1, 2, 3])

        # Check executions of the list comprehension
        for history, element in zip((0, 1, 2), num_list):
            expected_output = DataObject(
                hash=joblib.hash(element, hash_name='sha1'),
                hash_method="joblib_SHA1",
                type="numpy.float64", id=id(element),
                details={'shape': (), 'dtype': np.float64},
                value=element)

            _check_function_execution(
                actual=Provenance.history[history],
                exp_function=FunctionInfo('comprehension_function',
                                          'test_decorator', ''),
                exp_input={},
                exp_params={'param': history},
                exp_output={0: expected_output},
                exp_arg_map=['param'],
                exp_kwarg_map=[],
                exp_code_stmnt="num_list = [comprehension_function(i) for i in range(3)]",
                exp_return_targets=['num_list'],
                exp_order=1+history,
                test_case=self)

        # Check executions of the set comprehension
        for history, element in zip((3, 4, 5), num_set):
            expected_output = DataObject(
                hash=joblib.hash(element, hash_name='sha1'),
                hash_method="joblib_SHA1",
                type="numpy.float64", id=id(element),
                details={'shape': (), 'dtype': np.float64},
                value=element)

            _check_function_execution(
                actual=Provenance.history[history],
                exp_function=FunctionInfo('comprehension_function',
                                          'test_decorator', ''),
                exp_input={},
                exp_params={'param': history},
                exp_output={0: expected_output},
                exp_arg_map=['param'],
                exp_kwarg_map=[],
                exp_code_stmnt="num_set = {comprehension_function(i) for i in range(3, 6)}",
                exp_return_targets=['num_set'],
                exp_order=1+history,
                test_case=self)

        # Check executions of the dict comprehension
        for history, element in zip((6, 7, 8), num_dict.values()):
            expected_output = DataObject(
                hash=joblib.hash(element, hash_name='sha1'),
                hash_method="joblib_SHA1",
                type="numpy.float64", id=id(element),
                details={'shape': (), 'dtype': np.float64},
                value=element)

            _check_function_execution(
                actual=Provenance.history[history],
                exp_function=FunctionInfo('comprehension_function',
                                          'test_decorator', ''),
                exp_input={},
                exp_params={'param': history},
                exp_output={0: expected_output},
                exp_arg_map=['param'],
                exp_kwarg_map=[],
                exp_code_stmnt="num_dict = {i: comprehension_function(v) for i, v in enumerate(range(6, 9), start=1)}",
                exp_return_targets=['num_dict'],
                exp_order=1+history,
                test_case=self)


@Provenance(inputs=None, file_input=['file_name'])
def extract_words_from_file(file_name):
    with open(file_name, "r") as input_file:
        words = input_file.read().split(" ")
    return words


@Provenance(inputs=['words'], file_output=['file_name'])
def save_words_to_file(words, file_name):
    with open(file_name, "w") as output_file:
        output_file.writelines(" ".join(words))


class ProvenanceDecoratorFileInputOutputTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.res_path = Path(__file__).parent.absolute() / "res"

    def test_file_input(self):
        activate(clear=True)
        file_name = self.res_path / "file_input.txt"
        res = extract_words_from_file(file_name)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_list = ["This", "is", "an", "example", "file", "used", "as",
                         "input", "for", "the", "tests."]
        expected_output = DataObject(
            hash=joblib.hash(expected_list, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="builtins.list", id=id(res), details={}, value=None)

        expected_file = File("96ccc1380e069667069acecea3e2ab559441657807e0a86d14f49028710ddb3a",
                             hash_type="sha256", path=file_name)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('extract_words_from_file',
                                      'test_decorator', ''),
            exp_input={'file_name': expected_file},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['file_name'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = extract_words_from_file(file_name)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_file_output(self):
        activate(clear=True)
        tmp_file = tempfile.NamedTemporaryFile(dir=self.res_path, delete=True)
        file_name = tmp_file.name
        input_list = ["Some", "words", "were", "written", "to", "this", "file"]
        res = save_words_to_file(input_list, file_name)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_input = DataObject(
            hash=joblib.hash(input_list, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="builtins.list", id=id(input_list), details={}, value=None)

        # As None has its own UUID, let's get what was generated
        self.assertEqual(len(Provenance.history), 1)
        output_uuid = Provenance.history[0].output[0].hash

        expected_none_output = DataObject(hash=output_uuid, hash_method="UUID",
            type="builtins.NoneType", id=id(res), details={}, value=None)

        expected_file = File("00d20b4831b0dadded2c633bdfc3dde3926fc17baaed51dacdab3e52a3b0d419",
                             hash_type="sha256", path=Path(file_name))

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('save_words_to_file',
                                      'test_decorator', ''),
            exp_input={'words': expected_input},
            exp_params={},
            exp_output={0: expected_none_output, 'file.0': expected_file},
            exp_arg_map=['words', 'file_name'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = save_words_to_file(input_list, file_name)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)


# Tracking methods inside classes
class ObjectWithMethod(object):
    def __init__(self, coefficient):
        self.coefficient = coefficient

    def process(self, array, param1, param2):
        return array + self.coefficient

    @staticmethod
    def static_method(array, coefficient):
        return array + coefficient


ObjectWithMethod.process = Provenance(inputs=['self', 'array'])(
    ObjectWithMethod.process)

ObjectWithMethod.static_method = Provenance(inputs=['array'])(
    ObjectWithMethod.static_method)

# Apply decorator to method that uses the descriptor protocol
neo.AnalogSignal.reshape = Provenance(inputs=[0])(neo.AnalogSignal.reshape)


ObjectWithMethod.__init__ = Provenance(inputs=[])(ObjectWithMethod.__init__)


class ProvenanceDecoratorClassMethodsTestCase(unittest.TestCase):

    def test_static_method(self):
        obj = ObjectWithMethod(2)
        activate(clear=True)
        res = obj.static_method(TEST_ARRAY, 4)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        obj_info = DataObject(
            hash=joblib.hash(obj, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="test_decorator.ObjectWithMethod",
            id=id(obj),
            details={'coefficient': 2},
            value=None)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+4, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64},
            value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('ObjectWithMethod.static_method',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'coefficient': 4},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'coefficient'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = obj.static_method(TEST_ARRAY, 4)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_method_descriptor(self):
        activate(clear=True)
        ansig = neo.AnalogSignal(TEST_ARRAY, units='mV', sampling_rate=1*pq.Hz)
        reshaped = ansig.reshape((1, -1))
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_input = DataObject(
            hash=joblib.hash(ansig, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="neo.core.analogsignal.AnalogSignal", id=id(ansig),
            details={'_dimensionality': pq.mV.dimensionality,
                     '_t_start': 0 * pq.s, '_sampling_rate': 1 * pq.Hz,
                     'annotations': {}, 'array_annotations': {}, 'name': None,
                     'file_origin': None, 'description': None, 'segment': None,
                     'units': pq.mV.units, 'shape': (3, 1), 'dtype': np.int64,
                     't_start': 0 * pq.s, 't_stop': 3 * pq.s,
                     'dimensionality': pq.mV.dimensionality},
            value=None)

        expected_output = DataObject(
            hash=joblib.hash(reshaped, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="neo.core.analogsignal.AnalogSignal", id=id(reshaped),
            details={'_dimensionality': pq.mV.dimensionality,
                     '_t_start': 0 * pq.s, '_sampling_rate': 1 * pq.Hz,
                     'annotations': {}, 'array_annotations': {}, 'name': None,
                     'file_origin': None, 'description': None, 'segment': None,
                     'units': pq.mV.units, 'shape': (1, 3), 'dtype': np.int64,
                     't_start': 0 * pq.s, 't_stop': 1 * pq.s,
                     'dimensionality': pq.mV.dimensionality},
            value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('ndarray.reshape',
                                      'numpy', np.__version__),
            exp_input={0: expected_input},
            exp_params={1: (1, -1)},
            exp_output={0: expected_output},
            exp_arg_map=[0, 1],
            exp_kwarg_map=[],
            exp_code_stmnt="reshaped = ansig.reshape((1, -1))",
            exp_return_targets=['reshaped'],
            exp_order=1,
            test_case=self)

    def test_class_constructor(self):
        activate(clear=True)
        obj = ObjectWithMethod(2)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(obj, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="test_decorator.ObjectWithMethod",
            id=id(obj),
            details={'coefficient': 2},
            value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('ObjectWithMethod.__init__',
                                      'test_decorator', ''),
            exp_input={},
            exp_params={'coefficient': 2},
            exp_output={0: expected_output},
            exp_arg_map=['self', 'coefficient'],
            exp_kwarg_map=[],
            exp_code_stmnt="obj = ObjectWithMethod(2)",
            exp_return_targets=['obj'],
            exp_order=1,
            test_case=self)

    def test_object_method(self):
        obj = ObjectWithMethod(2)
        activate(clear=True)
        res = obj.process(TEST_ARRAY, 4, 5)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        obj_info = DataObject(
            hash=joblib.hash(obj, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="test_decorator.ObjectWithMethod",
            id=id(obj),
            details={'coefficient': 2}, value=None)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+2, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('ObjectWithMethod.process',
                                      'test_decorator', ''),
            exp_input={'self': obj_info, 'array': TEST_ARRAY_INFO},
            exp_params={'param1': 4, 'param2': 5},
            exp_output={0: expected_output},
            exp_arg_map=['self', 'array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = obj.process(TEST_ARRAY, 4, 5)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_class_constructor_container_output(self):
        activate(clear=True)
        obj = NonIterableContainerOutputObject(2)
        deactivate()

        self.assertEqual(len(Provenance.history), 4)

        elements = []
        for element in obj:
                element_info = DataObject(
                    hash=joblib.hash(element, hash_name="sha1"),
                    hash_method="joblib_SHA1",
                    type="numpy.int64", id=None,
                    details={'shape': (), 'dtype': np.int64},
                    value=element)
                elements.append(element_info)

        expected_output = DataObject(
            hash=joblib.hash(obj, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="test_decorator.NonIterableContainerOutputObject",
            id=id(obj),
            details={'_data': obj._data},
            value=None)

        # Check subscript of each element with respect to the container
        for history_index in (0, 1, 2):
            element = elements[history_index]
            _check_function_execution(
                actual=Provenance.history[history_index],
                exp_function=FunctionInfo('subscript', '', ''),
                exp_input={0: expected_output},
                exp_params={'index': history_index},
                exp_output={0: element},
                exp_arg_map=None,
                exp_kwarg_map=None,
                exp_code_stmnt=None,
                exp_return_targets=[],
                exp_order=None,
                test_case=self)

        _check_function_execution(
            actual=Provenance.history[3],
            exp_function=FunctionInfo(
                'NonIterableContainerOutputObject.__init__',
                                      'test_decorator', ''),
            exp_input={},
            exp_params={'start': 2},
            exp_output={0: expected_output},
            exp_arg_map=['self', 'start'],
            exp_kwarg_map=[],
            exp_code_stmnt="obj = NonIterableContainerOutputObject(2)",
            exp_return_targets=['obj'],
            exp_order=1,
            test_case=self)


@Provenance(inputs=['source'])
def use_dict(source):
    return 3


class ProvenanceDecoratorStoreValuesTestCase(unittest.TestCase):

    def setUp(self):
        alpaca_setting('store_values', [])

    def test_capture_dict(self):
        # This should have values for both the input dictionary and the
        # integer return
        alpaca_setting('store_values', ['builtins.dict'])
        activate(clear=True)
        test_dict = dict(id=[1, 2, 3], value={4, 5, 6})
        res = use_dict(test_dict)
        deactivate()

        dict_info = DataObject(hash=joblib.hash(test_dict, hash_name='sha1'),
                               hash_method="joblib_SHA1",
                               type="builtins.dict", id=id(test_dict),
                               details={},
                               value="{'id': [1, 2, 3], 'value': {4, 5, 6}}")

        expected_output = DataObject(hash=joblib.hash(3, hash_name='sha1'),
                                     hash_method="joblib_SHA1",
                                     type="builtins.int", id=id(res),
                                     details={}, value=3)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('use_dict', 'test_decorator', ''),
            exp_input={'source': dict_info},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['source'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = use_dict(test_dict)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_capture_builtins_only(self):
        # This should have values only for the integer return
        activate(clear=True)
        test_dict = dict(id=[1, 2, 3], value={4, 5, 6})
        res = use_dict(test_dict)
        deactivate()

        dict_info = DataObject(hash=joblib.hash(test_dict, hash_name='sha1'),
                               hash_method="joblib_SHA1",
                               type="builtins.dict", id=id(test_dict),
                               details={}, value=None)

        expected_output = DataObject(hash=joblib.hash(3, hash_name='sha1'),
                                     hash_method="joblib_SHA1",
                                     type="builtins.int", id=id(res),
                                     details={}, value=3)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('use_dict', 'test_decorator', ''),
            exp_input={'source': dict_info},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['source'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = use_dict(test_dict)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)


if __name__ == "__main__":
    unittest.main()



"""
Module to handle the serialization of Neo objects. This defines plugin
functions that are used by Alpaca when serializing information from Neo
objects in the captured provenance.

As Neo provides a specific data model for electrophysiology data, some
attributes need special handling. The key component is the `annotations`
dictionary, that cannot be stored as a standard Python attribute as the
information would not be accessible. A special property `hasAnnotation` is
used for those cases, where each key-value pair in the annotations dictionary
is identified as object metadata.

A special converter for attribute values is also provided, so that they can
be properly serialized to strings.
"""

from alpaca.ontology import ALPACA


__all__ = ['_neo_to_prov', '_neo_object_metadata']


NEO_COLLECTIONS = ('segments', 'events', 'analogsignals',
                   'spiketrains', 'channel_indexes', 'block',
                   'segment', 'epochs', 'parent', '_items', 'waveforms')

DISPLAYED_ATTRIBUTES = ('t_start', 't_stop', 'shape', 'dtype',
                        'name', 'description', 'nix_name')


def _neo_to_prov(value, displayed_attributes=DISPLAYED_ATTRIBUTES):
    # For Neo objects, we create a lightweight representation as a string, to
    # avoid dumping all the information such as SpikeTrain timestamps and
    # Event times as the usual Neo string representation. `value` is a Neo
    # object, and `displayed_attributes` is a list with the Neo object
    # attributes to be displayed in the result string.

    from alpaca.serialization.converters import _ensure_type

    type_information = type(value)
    neo_class = f"{type_information.__module__}.{type_information.__name__}"

    attr_repr = []
    for attribute in displayed_attributes:
        if hasattr(value, attribute):
            attr_value = getattr(value, attribute)
            attr_repr.append(f"{attribute}={_ensure_type(attr_value)}")

    neo_repr = f"{neo_class}({', '.join(attr_repr)})"
    return neo_repr


def _neo_object_metadata(graph, uri, metadata):
    # Adds metadata of a Neo object to an entity in the RDF graph `graph`.
    # `uri` is the identifier of the object in the graph, and `metadata` is
    # the dictionary of object metadata captured by Alpaca.
    # Returns a dictionary of attribute/annotation names with blank node
    # URIs, that are used later for inserting semantic information if
    # ontology annotations are defined.

    from alpaca.serialization.converters import _ensure_type
    from alpaca.serialization.prov import _add_name_value_pair

    metadata_nodes = {'attributes': {}, 'annotations': {}}

    for name, value in metadata.items():

        if name in NEO_COLLECTIONS:
            # A collection or a container...

            if isinstance(value, list):
                # This is a collection of Neo objects. Extract the
                # readable name of each Neo object in the list. They will be
                # enclosed in brackets [] as the value stored in the
                # serialized file.
                attr_values = []
                for item in value:
                    attr_values.append(_neo_to_prov(item))
                attr_value = f"[{', '.join(attr_values)}]"
            else:
                # This is a container Neo object. Just get the readable
                # name of the object
                attr_value = _neo_to_prov(value)

            # Add the attribute relationship to the object Entity
            blank_node = _add_name_value_pair(graph,
                                              uri=uri,
                                              predicate=ALPACA.hasAttribute,
                                              name=name,
                                              value=attr_value)
            metadata_nodes['attributes'][name] = blank_node

        elif name in ('annotations', 'array_annotations') and \
                isinstance(value, dict):
            # Handle the annotations in Neo objects

            for annotation, annotation_value in value.items():
                # Make sure that types such as list and Quantity are
                # handled
                annotation_value = _ensure_type(annotation_value)

                # Add the annotation relationship
                blank_node = _add_name_value_pair(graph,
                                            uri=uri,
                                            predicate=ALPACA.hasAnnotation,
                                            name=annotation,
                                            value=annotation_value)
                metadata_nodes['annotations'][name] = blank_node

        else:
            # Other attributes, just add them
            value = _ensure_type(value)

            # Add attribute relationship
            blank_node = _add_name_value_pair(graph,
                                              uri=uri,
                                              predicate=ALPACA.hasAttribute,
                                              name=name,
                                              value=value)
            metadata_nodes['attributes'][name] = blank_node

    return metadata_nodes



"""
This class provides functionality to serialize/deserialize the provenance
using an ontology based on the W3C Provenance Ontology (PROV-O). The Alpaca
ontology is used to serialize the provenance information captured by Alpaca as
RDF files.

.. autoclass:: alpaca.AlpacaProvDocument
    :members:

"""

from itertools import product, chain
import numpy as np
import numbers

from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF, PROV, XSD

from alpaca.ontology import ALPACA
from alpaca.serialization.identifiers import (data_object_identifier,
                                              file_identifier,
                                              function_identifier,
                                              script_identifier,
                                              execution_identifier,
                                              _get_function_name)
from alpaca.serialization.converters import _ensure_type
from alpaca.serialization.neo import _neo_object_metadata

from alpaca.utils.files import _get_prov_file_format
from alpaca.alpaca_types import DataObject, File, Container
from alpaca.settings import _ALPACA_SETTINGS
from alpaca.ontology.annotation import _OntologyInformation, ONTOLOGY_INFORMATION

from tqdm import tqdm


def _add_name_value_pair(graph, uri, predicate, name, value):
    # Add a relationship defined by `predicate` using a blank node as object.
    # The object will be of type `alpaca:NameValuePair`.
    blank_node = BNode()
    graph.add((uri, predicate, blank_node))
    graph.add((blank_node, RDF.type, ALPACA.NameValuePair))
    graph.add((blank_node, ALPACA.pairName, Literal(name)))
    graph.add((blank_node, ALPACA.pairValue, Literal(value)))
    return blank_node


class AlpacaProvDocument(object):
    """
    Generates a file using the Alpaca ontology (based on W3C PROV-O) from
    the history records captured during the execution of a Python script,
    or reads a serialized file into an RDF graph object.

    Attributes
    ----------
    graph : rdflib.Graph
        Provenance data represented as an RDF graph, using the Alpaca ontology
        based on PROV-O.

    Notes
    -----
    For convenience, you can serialize the active history easily by just using
    the :func:`save_provenance` in :ref:`interface`. This class should
    be used only if you want to access the data as an RDF graph or to manually
    control the serialization.
    """

    XSD_TYPES = {
        numbers.Integral: XSD.integer,
        numbers.Real: XSD.double,
        numbers.Complex: XSD.string,
        str: XSD.string,
        bool: XSD.boolean,
    }

    def __init__(self):
        self.graph = Graph()
        namespace_manager = self.graph.namespace_manager
        namespace_manager.bind('alpaca', ALPACA)
        namespace_manager.bind('prov', PROV)
        self._authority = _ALPACA_SETTINGS['authority']

        # Gets all OntologyInformation objects generated with annotation
        # information during the run. Update the current graph namespaces
        # accordingly
        _OntologyInformation.bind_namespaces(namespace_manager)

        # Metadata plugins are used for packages (e.g., Neo) that require
        # special handling of metadata when adding to the PROV records.
        # Plugins are external functions that take the graph, the object URI,
        # and the metadata dict as parameters. The function should return a
        # dictionary mapping all blank nodes generated to represent attributes
        # and annotations, to allow the use of any semantic information
        # defined by ontology annotations (i.e., __ontology__ attribute).
        self._metadata_plugins = {
            'neo': _neo_object_metadata
        }

        # Set to store all entity URIs that are added to the graph, so that
        # there is a fast lookup
        self._entity_uris = set()

        # Store functions that have container output ontology annotations,
        # To add the identification to the objects after the graph is built
        self._container_output_functions = {}
        for obj_type, info in ONTOLOGY_INFORMATION.items():
            container_returns = info.get_container_returns()
            if container_returns:
                self._container_output_functions[obj_type] = container_returns

    # PROV relationships methods

    def _wasAttributedTo(self, entity, agent):
        self.graph.add((entity, PROV.wasAttributedTo, agent))

    def _wasAssociatedWith(self, activity, agent):
        self.graph.add((activity, PROV.wasAssociatedWith, agent))

    def _wasDerivedFrom(self, used_entity, generated_entity):
        self.graph.add((generated_entity, PROV.wasDerivedFrom, used_entity))

    def _wasGeneratedBy(self, entity, activity):
        self.graph.add((entity, PROV.wasGeneratedBy, activity))

    def _used(self, activity, entity):
        self.graph.add((activity, PROV.used, entity))

    # Agent methods

    def _add_ScriptAgent(self, script_info, session_id):
        # Adds a ScriptAgent record from the Alpaca PROV model
        uri = URIRef(script_identifier(script_info, session_id,
                                       self._authority))
        self.graph.add((uri, RDF.type, ALPACA.ScriptAgent))
        self.graph.add((uri, ALPACA.scriptPath, Literal(script_info.path)))
        return uri

    # Activity methods

    def _add_Function(self, function_info):
        # Adds a Function record from the Alpaca PROV model
        uri = URIRef(function_identifier(function_info, self._authority))
        self.graph.add((uri, RDF.type, ALPACA.Function))
        self.graph.add((uri, ALPACA.functionName,
                        Literal(function_info.name)))
        self.graph.add((uri, ALPACA.implementedIn,
                        Literal(function_info.module)))
        self.graph.add((uri, ALPACA.functionVersion,
                        Literal(function_info.version)))
        return uri

    def _add_ontology_information(self, target_uri, ontology_info,
                                  information_type, element=None):
        class_info = ontology_info.get_uri(information_type, element)
        if class_info:
            if isinstance(class_info, list):
                for class_uri in class_info:
                    self.graph.add((target_uri, RDF.type, class_uri))
            else:
                self.graph.add((target_uri, RDF.type, class_info))

    def _add_FunctionExecution(self, script_info, session_id, execution_id,
                               function_info, params, execution_order,
                               code_statement, start, end, function,
                               ontology_info=None):
        # Adds a FunctionExecution record from the Alpaca PROV model
        uri = URIRef(execution_identifier(
            script_info, function_info, session_id, execution_id,
            self._authority))
        self.graph.add((uri, RDF.type, ALPACA.FunctionExecution))

        if ontology_info:
            self._add_ontology_information(uri, ontology_info, 'function')

        self.graph.add((uri, PROV.startedAtTime,
                        Literal(start, datatype=XSD.dateTime)))
        self.graph.add((uri, PROV.endedAtTime,
                        Literal(end, datatype=XSD.dateTime)))
        self.graph.add((uri, ALPACA.codeStatement, Literal(code_statement)))
        self.graph.add((uri, ALPACA.executionOrder,
                        Literal(execution_order, datatype=XSD.integer)))
        self.graph.add((uri, ALPACA.usedFunction, function))

        for name, value in params.items():
            value = _ensure_type(value)
            parameter_node = _add_name_value_pair(self.graph, uri,
                                                  ALPACA.hasParameter,
                                                  name, value)
            if ontology_info:
                self._add_ontology_information(parameter_node,
                                               ontology_info, 'arguments',
                                               name)
        return uri

    # Entity methods
    @classmethod
    def _get_entity_value_datatype(cls, info):
        value = info.value
        if value is None:
            return None

        # Check if builtin type or NumPy dtype
        value_class = value.__class__ if not isinstance(value, np.number) \
            else value.dtype.type
        if value_class in cls.XSD_TYPES:
            return cls.XSD_TYPES[value_class]

        # Check if object is include in the `store_values` setting.
        # In this case, they are always stored as strings
        obj_type = info.type
        if obj_type in _ALPACA_SETTINGS['store_values']:
            return XSD.string

        for possible_type in (numbers.Integral, numbers.Real, numbers.Complex):
            if issubclass(value_class, possible_type):
                return cls.XSD_TYPES[possible_type]

        # Type not found
        return None

    def _add_DataObjectEntity(self, info):
        # Adds a DataObjectEntity from the Alpaca PROV model
        # If the entity already exists, skip it
        uri = URIRef(data_object_identifier(info, self._authority))

        if uri in self._entity_uris:
            return uri

        self.graph.add((uri, RDF.type, ALPACA.DataObjectEntity))
        self.graph.add((uri, ALPACA.hashSource, Literal(info.hash_method)))

        value_datatype = self._get_entity_value_datatype(info)
        if value_datatype:
            self.graph.add((uri, PROV.value,
                            Literal(info.value, datatype=value_datatype)))

        ontology_info = ONTOLOGY_INFORMATION.get(info.type, None)
        if ontology_info:
            self._add_ontology_information(uri, ontology_info, 'data_object')

        self._add_entity_metadata(uri, info, ontology_info)
        self._entity_uris.add(uri)
        return uri

    def _add_FileEntity(self, info):
        # Adds a FileEntity from the Alpaca PROV model
        uri = URIRef(file_identifier(info, self._authority))
        self.graph.add((uri, RDF.type, ALPACA.FileEntity))
        self.graph.add((uri, ALPACA.filePath,
                        Literal(info.path, datatype=XSD.string)))
        return uri

    def _add_entity_metadata(self, uri, info, ontology_info=None):
        # Add data object metadata (attributes, annotations) to the entities,
        # using properties from the Alpaca PROV model
        package_name = info.type.split(".")[0]
        metadata = info.details

        if package_name in self._metadata_plugins:
            # Handle objects like Neo objects (i.e., to avoid dumping all the
            # information in collections such as `segments` or `events`)
            metadata_nodes = self._metadata_plugins[package_name](
                self.graph, uri, metadata)

            # Process metadata nodes of the object, if ontology information
            # defined
            if ontology_info:
                for metadata_type, elements in metadata_nodes.items():
                    for element, node in elements.items():
                        self._add_ontology_information(node, ontology_info,
                                                       metadata_type, element)
        else:
            # Add metadata using default handling, i.e., all attributes
            for name, value in metadata.items():
                # Make sure that types such as list and Quantity are handled
                value = _ensure_type(value)

                blank_node = _add_name_value_pair(self.graph, uri=uri,
                    predicate=ALPACA.hasAttribute, name=name, value=value)

                if ontology_info:
                    self._add_ontology_information(blank_node, ontology_info,
                                                   'attributes', name)

    def _add_membership(self, container, child, params):
        # Add membership relationships according to the standard PROV model
        # and properties specific to the Alpaca PROV model
        predicates = {
            'name': ALPACA.fromAttribute,
            'index': ALPACA.containerIndex,
            'slice': ALPACA.containerSlice,
        }

        for name, value in params.items():
            predicate = predicates[name]
            self.graph.add((child, predicate, Literal(value)))
        self.graph.add((container, PROV.hadMember, child))

    def _create_entity(self, info):
        # Create an Alpaca PROV Entity based on DataObject/File information
        if isinstance(info, DataObject):
            return self._add_DataObjectEntity(info)
        elif isinstance(info, File):
            return self._add_FileEntity(info)
        raise ValueError("Invalid entity!")

    # Interface methods

    def _add_function_execution(self, execution, script_agent, script_info,
                                session_id):
        # Add one `FunctionExecution` record to the file, and generate all the
        # provenance semantic relationships

        def _is_membership(function_info):
            name = function_info.name
            return name in ("attribute", "subscript")

        function_info = execution.function
        if _is_membership(function_info):
            # attributes and subscripting operations
            container = execution.input[0]
            child = execution.output[0]
            container_entity = self._create_entity(container)
            if PROV.wasAttributedTo not in \
                    self.graph.predicates(container_entity, script_agent):
                self._wasAttributedTo(container_entity, script_agent)
            child_entity = self._create_entity(child)
            self._add_membership(container_entity, child_entity,
                                 execution.params)
        else:
            # This is a function execution. Add Function activity
            cur_function = self._add_Function(function_info)

            # ID to identify ontology annotations
            info_id = _get_function_name(function_info)
            ontology_info = ONTOLOGY_INFORMATION.get(info_id)

            # Get the FunctionExecution node with function parameters and
            # other provenance info
            cur_activity = self._add_FunctionExecution(
                script_info=script_info, session_id=session_id,
                execution_id=execution.execution_id,
                function_info=function_info, params=execution.params,
                execution_order=execution.order,
                code_statement=execution.code_statement,
                start=execution.time_stamp_start,
                end=execution.time_stamp_end,
                function=cur_function, ontology_info=ontology_info
            )

            # Add all the inputs as entities, and create a `used` association
            # with the activity. URNs differ when the input is a file or
            # Python object.
            input_entities = []
            for key, value in execution.input.items():
                cur_entities = []
                has_input_uri = ontology_info and \
                                bool(ontology_info.get_uri('arguments', key))

                if isinstance(value, Container):
                    # If this is a Container, several objects are inside.
                    for element in value.elements:
                        cur_entities.append(self._create_entity(element))
                else:
                    cur_entities.append(self._create_entity(value))

                input_entities.extend(cur_entities)

                for cur_entity in cur_entities:
                    self._used(activity=cur_activity, entity=cur_entity)
                    self._wasAttributedTo(entity=cur_entity,
                                          agent=script_agent)
                    if has_input_uri:
                        self._add_ontology_information(cur_entity,
                                                       ontology_info,
                                                       'arguments', key)

            # Add all the outputs as entities, and create the `wasGenerated`
            # relationship.
            output_entities = []
            for key, value in execution.output.items():
                cur_entity = self._create_entity(value)
                output_entities.append(cur_entity)
                self._wasGeneratedBy(entity=cur_entity, activity=cur_activity)
                self._wasAttributedTo(entity=cur_entity, agent=script_agent)
                if ontology_info:
                    self._add_ontology_information(cur_entity, ontology_info,
                                                   'returns', element=key)

            # Iterate over the input/output pairs to add the `wasDerived`
            # relationship
            for input_entity, output_entity in \
                    product(input_entities, output_entities):
                self._wasDerivedFrom(used_entity=input_entity,
                                     generated_entity=output_entity)

            # Associate the activity to the script
            self._wasAssociatedWith(activity=cur_activity, agent=script_agent)

    def _add_annotations_for_container_outputs(self):
        # For functions that the Provenance decorator identified elements
        # inside returned containers, the elements linked by `prov:hasMember`
        # functions need to be annotated. The list of functions is already
        # stored in a search list. Iterate over the nodes of the function
        # and annotate the correct level of membership

        for info_id, levels in self._container_output_functions.items():

            # Initialize a container to store the URIs of elements of each
            # output level starting from the function. Since the capture can
            # ignore root levels, and to avoid recursion, we will map
            # container entities up to the maximum possible level taken from
            # the 'returns' annotations. Later, we take the annotations
            # starting from the deepest level.

            int_levels = list(map(lambda x: len(x), levels))
            max_level = max(int_levels)
            elements_by_level = {level: [] for level in range(max_level)}

            # Fetch information on the function, to identify nodes in the graph
            ontology_info = ONTOLOGY_INFORMATION[info_id]
            function_type = ontology_info.get_uri('function')
            executions = self.graph.subjects(RDF.type, function_type)

            # For every execution, get the output nodes
            # This is the first level
            for execution in executions:
                elements_by_level[0].extend(
                    self.graph.subjects(PROV.wasGeneratedBy, execution))

            # Traverse the remaining levels
            for level in range(1, max_level):
                for element in chain(elements_by_level[level-1]):
                    members = self.graph.objects(element, PROV.hadMember)
                    elements_by_level[level].extend(members)

            # Go from the deepest annotation level, annotating the deepest
            # node level with elements
            level_depth = max_level - 1
            level_str = '*' * max_level
            obj_uri = ontology_info.get_uri('returns', level_str)

            while level_depth >= 0:
                if obj_uri:
                    has_elements = False
                    for element in chain(elements_by_level[level_depth]):
                        has_elements = True
                        self.graph.add((element, RDF.type, obj_uri))
                else:
                    # No annotation requested for this level
                    # Consider the level traversed
                    has_elements = True

                if has_elements:
                    # Fetch annotation information for the parent level
                    level_str = '*' * (len(level_str) - 1)
                    obj_uri = ontology_info.get_uri('returns', level_str)

                # If no element found, keep the annotation level, but
                # try to annotate the elements of an upper node level
                level_depth -= 1

    def add_history(self, script_info, session_id, history,
                    show_progress=False):
        """
        Adds a history of `FunctionExecution` records captured by Alpaca to an
        RDF document using the Alpaca PROV ontology. The script is added as
        a `ScriptAgent` agent.

        Parameters
        ----------
        script_info : alpaca_types.File
            Named tuple with the information on the script being tracked
            (hash and file path).
        session_id : str
            Unique identifier for this script execution.
        history : list of FunctionExecution
            Provenance history to be serialized as RDF using PROV.
        show_progress : bool, optional
            If True, show the progress of the provenance history serialization.
            Default: False
        """
        script_agent = self._add_ScriptAgent(script_info, session_id)
        for execution in tqdm(history, desc="Serializing provenance history",
                              disable=not show_progress):
            self._add_function_execution(execution, script_agent, script_info,
                                         session_id)
        self._add_annotations_for_container_outputs()

    def read_records(self, file_name, file_format='turtle'):
        """
        Reads PROV data that was previously serialized as RDF.

        Parameters
        ----------
        file_name : str or Path-like
            Location of the file with PROV data to be read.
        file_format : {'json-ld', 'n3', 'nt', 'turtle', 'xml', 'ttl', 'rdf', 'json'}
            Format used to serialize the file that is being read. If None, the
            format will be inferred from the extension. The formats are
            the ones accepted by RDFLib. Some shortucts are defined for common
            file extensions:

            * 'ttl': Turtle
            * 'rdf': RDF-XML
            * 'json': JSON-LD

        Raises
        ------
        ValueError
            If `file_format` is None and `file_name` has no extension to infer
            the format, if it could not be inferred, or if the format is
            invalid.
        """
        if file_format is None:
            file_format = _get_prov_file_format(file_name)

        if file_format is None:
            raise ValueError("Could not infer serialization format. Please,"
                             "provide it explicitly using `file_format`.")

        if file_format not in ['json-ld', 'n3', 'nt', 'turtle', 'xml']:
            raise ValueError("Unsupported serialization format")

        with open(file_name, "r") as source:
            self.graph.parse(source, format=file_format)

    def serialize(self, file_name, file_format='turtle'):
        """
        Writes PROV data to a file or gets an in-memory string.

        Parameters
        ----------
        file_name : str or Path-like
            Location of the file with PROV data to be read.
        file_format : {'json-ld', 'n3', 'nt', 'hext', 'pretty-xml', 'trig', 'turtle', 'longturtle', 'xml'}
            Format used in the file that is being read. The format strings are
            the ones supported by RDFLib.
            If None, the format will be inferred from the extension.
            Default: 'turtle'

        """
        return self.graph.serialize(file_name, format=file_format)



import unittest

from alpaca import alpaca_setting
from alpaca.settings import _ALPACA_SETTINGS


class AlpacaSettingsTestCase(unittest.TestCase):

    def test_use_builtin_hash_for_module(self):
        setting_name = 'use_builtin_hash_for_module'

        cur_setting = alpaca_setting(setting_name)
        self.assertIsInstance(cur_setting, list)

        new_setting = alpaca_setting(setting_name, ['test'])
        self.assertListEqual(new_setting, ['test'])
        self.assertListEqual(alpaca_setting(setting_name), ['test'])
        self.assertListEqual(_ALPACA_SETTINGS[setting_name], ['test'])

        # Test wrong type
        with self.assertRaises(ValueError):
            alpaca_setting(setting_name, "test wrong type")

        # Restore value
        alpaca_setting(setting_name, cur_setting)
        self.assertListEqual(_ALPACA_SETTINGS[setting_name], cur_setting)

    def test_authority(self):
        setting_name = 'authority'

        cur_setting = alpaca_setting(setting_name)
        self.assertIsInstance(cur_setting, str)

        new_setting = alpaca_setting(setting_name, "test-authority")
        self.assertEqual(new_setting, "test-authority")
        self.assertEqual(alpaca_setting(setting_name), "test-authority")
        self.assertEqual(_ALPACA_SETTINGS[setting_name], "test-authority")

        # Test wrong type
        with self.assertRaises(ValueError):
            alpaca_setting(setting_name, ["test wrong type"])

        # Restore value
        alpaca_setting(setting_name, cur_setting)
        self.assertEqual(_ALPACA_SETTINGS[setting_name], cur_setting)

    def test_wrong_setting_name(self):
        with self.assertRaises(ValueError):
            alpaca_setting("wrong_setting")


if __name__ == "__main__":
    unittest.main()